//
//  Guit GUI Toolkit Examples
//  Copyright © 2020 Eric Lecolinet. All rights reserved.
//  http://www.telecom-paris.fr/~elc
//

#include <guit.hpp>
#include <gshape.hpp>
#include "examples.hpp"
#include <gcursor.hpp>
//#include <gmactouch.hpp>
using namespace std;
using namespace guit;


class DrawingCanvas : public GChoiceBox {
public:
  enum Action {TSelect, TLine, TRect, TOval, TFreeHand};

  DrawingCanvas();
  void erase();
  void eraseAll();
  void restore();

  GBool changed;
  GChoice action;
  GCheck multicreate;
  GColor color;
  GBackground bgcolor;
  GPos  pos;
  GSize size;
  GText info;

protected:
  void bindItem(Gadget* item);
  void press(GMouseEvent&);
  float minsize_{5.f};
  GBox undolist_;
};


class DrawingPalette : public GButtonBox {
public:
  DrawingPalette(DrawingCanvas&);
  void startToolglass(GWindow&);
  void start();
  void stop();

private:
  void initCB();
  void mouseDownCB(GMouseEvent&);
  DrawingCanvas& canvas;
};


class Drawing : public GWindow {
public:
  enum RemoteMode {StandAlone, Client, Server};
  
  Drawing() : GWindow("#drawing Drawing")
  {
    // shows an alert before opening the dialog if canvas.changed is true
    auto& opendial = OpenDialog("#open-dial",":/data")
    .alertIf(canvas.changed);


    // reads shapes and add them to the canvas
    opendial << [&] {
      canvas.read(opendial.pathname());
      canvas.changed = false;
    };


    auto& savedial = SaveDialog("#save-dial","~");
    savedial << [&] {
      canvas.write(savedial.pathname(), GIOMode::Content);
      canvas.changed = false;
    };


    auto& quitdial = QuitDialog("#quit-dial Close Window",
                                "Do you want to close the window?")
    // open dialog only if canvas.changed is true, otherwise just quit
    .openOnlyIf(canvas.changed);


    // controls whether the right panel is shown
    auto& show_right_panel = Check(true);

    auto& toolbar = ToolBar
    << (".open-btn" << opendial)
    << (".save-btn" << savedial)
    << Separator
    << (ListBox("#actions")
        << canvas.action
        << ".select-btn"
        << ".line-btn"
        << ".rect-btn"
        << ".oval-btn"
        << ".freehand-btn"
        )
    << (ChoiceItem(".multicreate-btn") << canvas.multicreate)
    << Separator
    << (".untrash-btn"  << [&]{canvas.restore();})
    << (".trash-btn"    << [&]{canvas.erase();})
    << Separator
    << (".trashall-btn" << [&]{canvas.eraseAll();})
    << Separator
    << (ChoiceItem(".palette-btn") << [&]{palette.show(!palette.isShown());})
    << (".overview-btn" << [&]{showOverview(canvas);})
    << ("Print"         << [&]{canvas.write(cout, GIOMode::Content);})
    // << ("Test" << [&]{mainwin.write(cout, Gadget::IOMode::Gadgets);})
    << Filler
    << (ChoiceItem(".settings-btn") << show_right_panel);


    *this
    << quitdial
    << palette
    << toolbar
    << (StatusBar << (Label << canvas.info))
    << (HBox
        // drawing area
        << (ScrollBox << canvas)
        // right panel (its visibility depends on show_right_panel)
        << (VBox(".right-panel")
            << (Show() <<= show_right_panel)
            << (VBox(".subpanel <b>Background") << makeSliders(canvas.bgcolor))
            << (VBox(".subpanel <b>Foreground") << makeSliders(canvas.color))
            )
        );

    canvas.action = 2;
    canvas.multicreate = true;
    palette.update();
    canvas.changed = false;
  }

private:
  DrawingCanvas canvas;
  DrawingPalette palette{canvas};


  Gadget& makeSliders(GColor& c) {
    // create sliders and bind their value with the corresponding color component
    return Table
    << makeSlider(".red-field", c.redField())
    << makeSlider(".green-field", c.greenField())
    << makeSlider(".blue-field", c.blueField())
    << makeSlider(".alpha-field", c.alphaField());
  }


  GBox& makeSlider(GString const& label, GColor::Field field) {
    auto& value = Float(0.f, 0.f, 1.f);  // alpha comp between 0 and 1
    value ^= field;   // ^= syncs value with field
    return HBox(".text-slider {align:x x; gap:0; margin:0}")
    << label
    << (Slider(".slider") << value)
    << (TextField(".text") << value);
  }


  void showOverview(DrawingCanvas& drawing) {
    static gptr<GDialog> overview;
    if (!overview) {
      overview = Dialog(".overview Overview")
      << (ScrollBox << drawing.clone(GClone::Sync));
      overview->setScale(0.5);
    }
    overview->show();
  }
};


GWindow& Examples::makeDrawing() {
  return *new Drawing;
}



DrawingCanvas::DrawingCanvas() : GChoiceBox(".canvas") {
  // register these shapes to allow creating them when reading a file
  // even if none of them have been yet created
  GRectShape::Type();
  GOvalShape::Type();
  GArcShape::Type();
  GLineShape::Type();
  GFreehandShape::Type();

  *this
  // called when the mouse is pressed on the canvas
  << on::mousedown / [&](GEvent& e){
    press(e.asMouseEvent());}
  // called when an item is selected
  << on::change / [&](GEvent& e){bindItem(e.chosen());};

  // init canvas attributes
  color = "0 0 0 1";
  bgcolor = "0.5 0.5 0.5 0.5";
  info = "Ready";

  // when not in TSelect (selection) mode, the canvas catches events
  // which allows drawing on the top of an existing shape
  action << [&]{catchEvents(action != TSelect);};
}

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

void DrawingCanvas::press(GMouseEvent& e) {
  info = "";
  choose(nullptr);
  bindItem(nullptr);

  GShape* shape{};

  switch (action) {
    case TSelect:
      return;
    case TRect:
      shape = new GRectShape();
      break;
    case TOval:
      shape = new GOvalShape();
      break;
    case TLine:
      shape  = new GLineShape();
      break;
    case TFreeHand:
      shape = new GFreehandShape();
      break;
  }

  if (shape) {
    *shape
    << new GDragPos(e.pos())
    << new GDragSize()
    << new GColor(color)
    << new GBackground(bgcolor);

    *this << shape;
    shape->reshapeInteractively(e);

    changed = true;
    if (!multicreate()) action = TSelect;
  }
}

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

void DrawingCanvas::bindItem(Gadget* item) {
  // when a new item is selected:
  // 1) unbind previous item
  // 2) bind new item if not null
  color.unbind(1);
  bgcolor.unbind(1);
  pos.unbind(1);
  size.unbind(1);

  if (!item) return;

  gsync(color, item->prop<GColor>(), 1);
  gsync(bgcolor, item->prop<GBackground>(), 1);
  gsync(pos, item->prop<GPos>(), 1);
  gsync(size, item->prop<GSize>(), 1);

  // move the item to the end to place it on the top of the other items
  if (childAt(-1) != item) {
    *this << *item;
    choose(item);  // must reselect the item because it has moved
  }
}

void DrawingCanvas::erase() {
  if (auto* item = chosen()) {
    undolist_ << item;    // so that we can restore the item later
    choose(nullptr);
    if (childCount() == 0) changed = false;
    update();
  }
}

void DrawingCanvas::eraseAll() {
  choose(nullptr);
  auto list = children();  // copy needed because the loop modifies chidren
  for (auto& item : list) {
    undolist_ << item;
  }
  changed = false;
  update();
}

void DrawingCanvas::restore() {
  // if an item at the end of the krestorelist, put it back to the canvas
  if (auto* item = undolist_.childAt(-1)) {
    *this << *item;
  }
}

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

DrawingPalette::DrawingPalette(DrawingCanvas& c)
: GButtonBox("#palette"), canvas(c) {
  *this
  << new GDragPos()
  << on::init / [this]{initCB();}
  << on::show / [this]{start();}
  << on::hide / [this]{stop();};
}

void DrawingPalette::start() {
  move(10,80);
  //GTouchManager::instance().start();
}

void DrawingPalette::stop() {
  //GTouchManager::instance().stop();
}

// toolglass mode: an external touchpad (or the left side of the embedded
// touchpad if there is no external touchpad) will move the palette
void DrawingPalette::initCB() {
  //cerr << "Palette::initCB" << endl;

  for (auto& btn : children()) {
    btn << on::mousedown / [this](GEvent& e){mouseDownCB(e.asMouseEvent());};
  }

  // starts direct touch detection, creates as many psudo-cursors as there are touchpads
  // 0 is the ID the native cursor, positive values corresponds to pseudo-cursors
  /*
  auto& tm = GTouchManager::instance();
  auto padcount = tm.init(*getWindow());  // returns the number of touchpads

  // if at least one touchpad
  if (padcount >= 1) {
    // if only 1 touchpad, make second finger to act as cursor(2)
    if (padcount == 1) tm.getPad(0)->createCursor2(GColor::blue);

    if (auto* curs2 = GCursor::cursor(2)) {
      // palette position is bound to cursor(2)
      prop<GPos>() <<= *curs2;
      // hide cursor(2)
      //curs2->show(false);
    }
  }*/
}

void DrawingPalette::mouseDownCB(GMouseEvent& e) {
  // get bgcolor of button
  auto* color = e.from().findProp<GBackground>();

  if (canvas.action == DrawingCanvas::TSelect) {
    // find shape on which the button was pressed and change its color
    auto* target = canvas.childAtWPoint(e.wpos());
    //cerr << "mouseDown " << color << " target " << target << endl;
    if (target) target->prop<GBackground>() = *color;
  }
  else {
    canvas.choose(nullptr);
    canvas.bgcolor = *color;
    e.redirectEvent(canvas);
  }
}
