//
// Core Layer & Interaction Controller
//
// Guit GUI Toolkit (https://www.telecom-paris.fr/~elc/guit/)
// Copyright 2020 Eric Lecolinet (eric.lecolinet@gmail.com) - Telecom Paris
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//

#ifndef Guit_Control_hpp
#define Guit_Control_hpp
#include <list>
#include <map>
#include <functional>
#include <gprefs.hpp>
#include <gadget.hpp>
#include <gcursor.hpp>
namespace guit {

#define GIMPLEMENT_GADGET_TYPE(T) \
G##T::GType& G##T::Type() {static auto* t = new G##T::GType(#T); return *t;} \

#define GIMPLEMENT_GADGET(T)  GIMPLEMENT_GADGET_TYPE(T) \
G##T& T(GString const& spec) {return *new G##T(spec);}

#define GIMPLEMENT_GADGET_ALL(T,Super)  GIMPLEMENT_GADGET(T) \
G##T::G##T(GString const& spec) : Super(Type(), spec) {} \
G##T& G##T::clone(GClone const& c) { \
  auto* g = new G##T; g->initClone(*this, c); return *g; \
}

#define GIMPLEMENT_PROP_TYPE(T,R) \
G##T::GType& G##T::Type() {static auto* t = new G##T::GType(#T,R); return *t;}

#define GIMPLEMENT_PROP(T,R) GIMPLEMENT_PROP_TYPE(T,R) \
G##T& T() {return *new G##T;} \
G##T& T(GString const& val) {return *new G##T(val);}


/// Fatal Error exception.
/// An error message is displayed and the main loop is restarted when such an exception is thrown.
class GFatalError {
public:
  GFatalError(GString const& where, GString const& message, GObject const* object = {});
  GString what_;
};


class GPickMode {
public:
  bool find_{}, catchedevent_{};
  Gadget *catchedchild_{};
  std::function<bool(Gadget&)> pred_{};
};


/// Input channel.
class GChannel : public GCursor {
public:
  GChannel(unsigned int id);

  struct Clicked {
    Gadget* target;  // last clicked
    GTime time;      // time when last clicked
    int count;       // click count
  };
  // gprotect avoids destructing object while processing events
  gprotect<Gadget>
  enterg_,   // last entered
  pressg_,   // last pressed
  focusg_,   // keyboard focus
  browseg_;  // last browsed item in menus
  Clicked clicked_;
  GEventState buttonstate{};
};

using GChannels = std::vector<GChannel*>;


class GHotkeys : public std::map<uint64_t, gptr<Gadget>> {};


// Guit Core.
class GCore : public GPrefs {
protected:
  GCore();
  GCore(GCore const&) = delete;
  GCore(GCore const&&) = delete;
  void operator=(GCore const&) = delete;
  void operator=(GCore const&&) = delete;

public:
  struct BeginPlainObjects {
    BeginPlainObjects() {GCore::setPlainMode(true);}
  };
  struct EndPlainObjects {
    EndPlainObjects() {GCore::setPlainMode(false);}
  };

  static int start(GWindow* mainwindow);
  static void exit(int status);
  static void subLoop(const bool& quit);
  static bool isRunning();
  static GTime elapsedTime();
  static class GNatContext* natApp();

  static GId getId(GString const& name, bool obtain);
  static GId findId(GString const& name) {return getId(name, false);}
  static GId obtainId(GString const& name) {return getId(name, true);}

  static GChannel* channel(unsigned int id);
  static GChannel& obtainChannel();
  static GChannel& obtainChannel(unsigned int id);
  static GChannels& channels();

  static void paintAll();
  static GWindow* mainWindow();
  static GWindow* modalWindow();
  static GStyle* mainWindowStyle();
  static std::list<GWindow*>& windowList();
  static void addWindow(GWindow&);
  static void removeWindow(GWindow&);
  static void setModalWindow(GWindow*);
  static void showOverlay(GBox& overlay, Gadget* opener,
                          bool state, bool autoclose, bool showaswin);
  static void updateLayout(GWindow* upd_win, Gadget* upd_root, bool upd_css);
  static void requestUpdate();

  static void printClassSizes();
  static bool plainMode();
  static void setPlainMode(bool mode);

  static int inspecting() {return inspecting_;}
  static Gadget* inspected() {return inspected_;}

  static int argc_;
  static const char** argv_;
  static class GTheme* theme_;
  static class GInspector* inspector_;
  static int  inspecting_;
  static Gadget* inspected_;
  static std::function<void(Gadget*)> doinspect_;
};

}
#endif
