//
//  Text Renderer
//
// Guit GUI Toolkit (https://www.telecom-paris.fr/~elc/guit/)
// Copyright 2020 Eric Lecolinet (eric.lecolinet@gmail.com) - Telecom Paris
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//

#include <gtextitem.hpp>
#include <gicon.hpp>
#include <gcontrol.hpp>
#include <grender.hpp>
#include <gtextrender.hpp>
#include <gnative.hpp>
#include <gstring.hpp>
using namespace std;
namespace guit {

GDim GTextRender::layoutText(GFontValue const& font, bool richtext,
                             GString const& text, float areawidth,
                             GTextAttributes const* attrs, GTextCache* cache) {
  font.realize();
  size_t begspan = 0, textlen = text.size();
  float curw = 0, maxw = 0;

  GTextCache tmpcache;
  if (!cache) cache = &tmpcache;
  if (!cache->rows_) cache->rows_ = new std::vector<GTextCache::Row>;
  else cache->rows_->clear();

  // add first row
  cache->rows_->push_back(GTextCache::Row(0, 0, 0));
  auto* lastrow = &cache->rows_->back();
  size_t k = 0;

  while (true) {

    if (k >= textlen) {      // last line
      return GDim(maxw, lastrow->y_ + lastrow->h_);
    }

    // find end of next span
    while (k < textlen && !::isspace(text[k])) k++;

    // get span size
    auto spansize = font.textSize(text.substr(begspan,k-begspan+1),
                                  richtext, attrs);

    if (curw + spansize.w <= areawidth) {  // add span on same line
      curw += spansize.w;
      if (curw > maxw) maxw = curw;
      if (lastrow->h_ < spansize.h) lastrow->h_ = spansize.h;
      lastrow->length_ = k - lastrow->start_ +1;
      begspan = k+1;

      // \n found at the end of text, must change line
      if (text[k] == '\n') {  // add empty line
        lastrow->length_--;   // remove \n
        cache->rows_->push_back(GTextCache::Row(lastrow->y_ + lastrow->h_, 0, 0));
        lastrow = &cache->rows_->back();
        // lastrow->h must not be 0,  the last line wont show otherwise
        lastrow->h_ = font.fontHeight();
        lastrow->start_ = k+1;
        lastrow->length_ = 0;
        curw = 0;
        begspan = k+1;
      }
    }
    else { // add line and put span on it
      cache->rows_->push_back(GTextCache::Row(lastrow->y_ + lastrow->h_, 0, 0));
      lastrow = &cache->rows_->back();
      curw = spansize.w;
      if (curw > maxw) maxw = curw;
      lastrow->h_ = spansize.h;
      lastrow->start_ = begspan;
      lastrow->length_ = k - lastrow->start_ + 1;
      begspan = k+1;

      if (text[k] == '\n') {  // add empty line
        lastrow->length_--;   // remove \n
        cache->rows_->push_back(GTextCache::Row(lastrow->y_ + lastrow->h_, 0, 0));
        lastrow = &cache->rows_->back();
        // lastrow->h must not be 0,  the last line wont show otherwise
        lastrow->h_ = font.fontHeight();
        lastrow->start_ = k+1;
        lastrow->length_ = 0;
        curw = 0;
        begspan = k+1;
      }
    }
    k++;
  }
}


GTextRender::GTextRender(GFontValue const& font, bool richtext, const GRgba* color,
                         size_t selbegin, size_t selend,
                         GTextAttributes const* attrs) :
htmltags_(richtext), pos_(0), begpart_(0)
{
  if (attrs && !attrs->list_.empty()) attrlist_ = attrs;
  else attrlist_ = nullptr;

  font.realize();

  if (color)
    stack_.emplace_back(StackNode(font, *color, GRgba(0.f,0.f,0.f,0.f)));
  else
    stack_.emplace_back(StackNode(font, GRgba(), GRgba(0.f,0.f,0.f,0.f)));

  current_ = &(stack_[0]);

  if (selbegin <= selend) {
    selbegin_ = selbegin;
    selend_ = selend;
  }
  else {
    selbegin_ = selend;
    selend_ = selbegin;
  }
}


bool GTextRender::findCurrentAttr(size_t offset) {
  if (curattrpos_ >= attrlist_->list_.size()) {
    curattr_ = nullptr;
    return false;
  }
  auto* a = attrlist_->list_[curattrpos_];
  if (a && (pos_+offset) <= a->pos_ + a->len_) {
    curattr_ = a;
    return true;
  }
  else {
    ++curattrpos_;
    if (curattrpos_ >= attrlist_->list_.size()) {
      curattr_ = nullptr;
      return false;
    }
    else {
      curattr_ = attrlist_->list_[curattrpos_];
      return true;
    }
  }
}

enum ATTR {
  NoAttr=0, BeginList=1, EndList=2, BeginAttr=3, BeginStyle=4, Skip=5
};

int GTextRender::attrFound(GString const& text, size_t offset) {
  if (attrlist_) {
    findCurrentAttr(offset);
    if (!curattr_) return 0;
    else if (pos_ + offset == curattr_->pos_) return BeginList;
    else if (pos_ + offset == curattr_->pos_ + curattr_->len_) return EndList;
    else return 0;  // inside start and end
  }
  else if (htmltags_) {
    if (text[pos_] == '<') return BeginAttr;
    else if (text[pos_] == '{') return BeginStyle;
    else if (text[pos_] == '\\' && pos_+1 < text.size()) return Skip;
    else return 0;
    /*
     if (text[pos_] != '<') return false;
     else if (pos_ == 0) return true;
     else return text[pos_-1] != '<';
     */
  }
  else return 0;
}


GDim GTextRender::getSize(GString const& text) {
  current_->font_.realize();

  if ((!htmltags_ && !attrlist_) || text.empty()) {
    textsize_ = current_->font_.textSize(text, false, attrlist_);
    return textsize_;
  }

  pos_ = begpart_ = 0;
  textsize_ = {};

  while (pos_ < text.size()) {
    int delim = attrFound(text, 0/*offset*/);
    if (delim == 0) ++pos_;
    else {
      if (pos_ > begpart_) {
        GString str = text.substr(begpart_, pos_ - begpart_);
        GDim strsize = current_->font_.textSize(str, false, nullptr);
        textsize_.w += strsize.w;
        textsize_.h = max(textsize_.h, strsize.h);
      }
      if (!processAttr(nullptr, 0, 0, delim, text, 0/*offset*/)) break;  // changes begpart_ and pos_
    }
  }

  if (begpart_ == 0) {
    textsize_ = current_->font_.textSize(text, false, nullptr);
  }
  else if (pos_ - begpart_ > 0) {
    GString str = text.substr(begpart_, pos_ - begpart_);
    GDim strsize = current_->font_.textSize(str, false, nullptr);
    textsize_.w += strsize.w;
    textsize_.h = max(textsize_.h, strsize.h);
  }
  return textsize_;
}


void GTextRender::draw(GGraphics& g, float x0, float y0,
                       GString const& text, size_t offset) {
  auto textlen = text.size();

  if (text.empty()) {
    current_->font_.realize();
    textsize_ = current_->font_.textSize(text, false, nullptr);
    return;
  }

  textsize_ = {};
  pos_ = begpart_ = 0;
  bool insel = (pos_+offset >= selbegin_) && (pos_+offset < selend_);

  while (true) {
    int delim = attrFound(text, offset);

    if (delim > 0) {
      if (pos_ - begpart_ > 0) drawPart(g, x0, y0, text, false, insel);

      insel = (pos_+offset >= selbegin_) && (pos_+offset < selend_);

      if (delim == Skip) {    // \\x just skip and dont process x
        ++pos_;
        begpart_ = pos_;
        ++pos_;
      }
      // note: changes begpart_ and pos_
      else if (!processAttr(&g, x0, y0, delim, text, offset)) break;
    }
    
    else if (pos_+offset == selbegin_ || pos_+offset == selend_ || pos_ >= textlen) {
      drawPart(g, x0, y0, text, delim == EndList, insel);

      insel = (pos_+offset >= selbegin_) && (pos_+offset < selend_);
      begpart_ = pos_;
      if (pos_ >= textlen) return;
      ++pos_;
    }
    
    else ++pos_;
  }
}


void GTextRender::drawPart(GGraphics& g, float x, float y,
                           GString const& text, bool drawlast, bool insel) {
  GString str;
  GDim strsize;

  if (curattr_ && drawlast) str = text.substr(begpart_, pos_ - begpart_ +1);
  else str = text.substr(begpart_, pos_ - begpart_);

  g.setFont(current_->font_);
  g.setColor(current_->color_);
  g.setBgColor(insel ? GPrefs::selectedTextColor() : current_->bgcolor_);
  strsize = g.drawString(x + textsize_.w, y, str);
  textsize_.w += strsize.w;
  textsize_.h += strsize.h;  // ???
}


// graphics is null if called by computeSize()
bool GTextRender::processAttr(GGraphics* g, float x, float y, int delim,
                              GString const& text, size_t offset) {

  if (attrlist_) {   // list of attributes
    if (!curattr_) return false;

    if (pos_ + offset == curattr_->pos_) {
      processAttrBegin(g, x, y, curattr_->props_);
      begpart_ = pos_;
      pos_++;
      return true;
    }
    else if (pos_ + offset == curattr_->pos_ + curattr_->len_) {
      processAttrEnd();
      begpart_ = pos_;
      pos_++;
      return true;
    }
    else return false;
  }

  else if (htmltags_) {   // pseudo-html text
    auto len = text.size();
    auto pos = pos_;

    ++pos;
    if (delim == BeginStyle) {
    while (pos < len && text[pos] != '}') ++pos;
      pos_ = pos + 1;
      begpart_ = pos_;
      return true;
    }

    while (pos < len && text[pos] != '>') ++pos;
    if (pos >= len) {
      pos_ = pos;
      return false;
    }
    if (text[pos_ + 1] != '/') {
      processAttrBegin(g, x, y, text.substr(pos_ + 1, pos - pos_ - 1));
    }
    else {
      processAttrEnd();
    }
    pos_ = pos + 1;
    begpart_ = pos_;
    return true;
  }

  else return false;
}


void GTextRender::processAttrBegin(GGraphics* g, float x, float y,
                                   GPropList const& props) {
  for (auto& it : props) {
    switch (it->role()) {

      case GPropType::Font:
        stack_.emplace_back(stack_.back());
        current_ = &(stack_.back());
        current_->font_ = static_cast<GFont*>(it.get())->value_;
        break;

      case GPropType::Color:
        stack_.emplace_back(stack_.back());
        current_ = &(stack_.back());
        current_->color_ = static_cast<GColor*>(it.get())->value_;
        break;

      case GPropType::Background:
        stack_.emplace_back(stack_.back());
        current_ = &(stack_.back());
        current_->bgcolor_ = static_cast<GBackground*>(it.get())->value_;
        break;
    }
  }
}


void GTextRender::processAttrBegin(GGraphics* g, float x, float y,
                                   GString const& attr) {
  if (attr.empty()) return;

  GString tag, value;
  size_t from{};
  // note: quotes not compatible with current version of GText::nextCharIndex()
  // " " needed because of <img src=...>
  int delim = strtoken(attr, tag, from, "=: ");  // trims
  if (delim < 0) return;

  if ((delim == '=' || delim == ':') && strtoken(attr, value, from) < 0)  // trims
    goto ERROR;

  switch (strid(tag)) {             // il manque <br> et <br/> !!!!
    case strid("i"):
    case strid("em"):     // pas vraiment correct !!!
      stack_.emplace_back(stack_.back());
      current_ = &(stack_.back());
      current_->font_ = GFont::italic.value_;
      return;

    case strid("b"):
      stack_.emplace_back(stack_.back());
      current_ = &(stack_.back());
      current_->font_ = GFont::bold.value_;
      return;

    case strid("c"):
    case strid("color"):
      if (!g) return;
      if (value.empty()) goto ERROR;
      stack_.emplace_back(stack_.back());
      current_ = &(stack_.back());
      current_->color_ = GRgba(value);
      return;

    case strid("bg"):
    case strid("background"):
      if (!g) return;
      if (value.empty()) goto ERROR;
      stack_.emplace_back(stack_.back());
      current_ = &(stack_.back());
      current_->bgcolor_ = GRgba(value);
      return;

    case strid("f"):
    case strid("font"):
      if (!g) return;
      if (value.empty()) goto ERROR;
      stack_.emplace_back(stack_.back());
      current_ = &(stack_.back());
      current_->font_ = GFontValue(value);
      return;

    case strid("small"):
      stack_.emplace_back(stack_.back());
      current_ = &(stack_.back());
      current_->font_ = GFont::small.value_;
      return;

    case strid("big"):
    case strid("large"):
      stack_.emplace_back(stack_.back());
      current_ = &(stack_.back());
      current_->font_ = GFont::large.value_;
      return;

    case strid("img"):
      processImage(g, x, y, from, attr);
      return;
  }
ERROR:
  // trop bavard!! cf vrrbose mode
  // gerror("GTextRender","missing value in <"+attr+">");
  return;
}


void GTextRender::processAttrEnd() {
  if (stack_.size() > 1) stack_.pop_back();
  current_ = &(stack_.back());
}


// graphics is null if called by computeSize()
void GTextRender::processImage(GGraphics* g, float x, float y,
                               size_t from, GString const& attr) {
  GStrings tab;
  size_t n = strsplit(attr, tab, "=: ");
  float w{-1}, h{-1};
  GString src;

  for (size_t k = 1; k < n; k += 2) {   // skip "img"
    switch (strid(tab[k])) {
      case strid("src"):
        if (k+1 < n) src = tab[k+1];
        break;

      case strid("w"):
      case strid("width"):
        if (k+1 < n) w = float(::atoi(tab[k+1].data()));
        break;

      case strid("h"):
      case strid("height"):
        if (k+1 < n) h = float(::atoi(tab[k+1].data()));
        break;
    }
  }
  if (src.empty()) {
    gerror("GTextRender","missing src value in <"+attr+">");
    return;
  }
  GIcon ima(src);
  if (!GGraphics::readImage(ima.value_)) {
    gerror("GTextRender","could not find image in <"+attr+">");
    return;
  }

  if (w < 0 && h < 0) {
    w = float(ima.value_.w);
    h = float(ima.value_.h);
  }
  else if (w >= 0) {
    h = ima.value_.h * w / ima.value_.w;
  }
  else if (h >= 0) {
    w = ima.value_.w * h / ima.value_.h;
  }
  // y is text lower base
  if (g) g->drawImage(x + textsize_.w, y - h, w, h, ima.value_);
  textsize_.w += w;
  textsize_.h += h;  // ???
}

}
