//
//  Layout of container gadgets
//
// Guit GUI Toolkit (https://www.telecom-paris.fr/~elc/guit/)
// Copyright 2020 Eric Lecolinet (eric.lecolinet@gmail.com) - Telecom Paris
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//

#include <algorithm>
#include <gevent.hpp>
#include <gborder.hpp>
#include <gbox.hpp>
#include <gvalueitem.hpp>
#include <gcontrol.hpp>
#include <grender.hpp>
#include <gstyle.hpp>
using namespace std;
namespace guit {

static bool isGrid(GFlow::ValueType const& val)  {
  return val.gridsize >= 1 && val.kind == GFlow::Row;
}

GRect GBox::childrenArea(GRender& r) const {
  GRect area = frame_;
  if (r.margin_) {
    area.x += r.margin_->left;
    area.y += r.margin_->top;
    area.w -= r.margin_->left + r.margin_->right;
    area.h -= r.margin_->top  + r.margin_->bottom;
  }
  if (r.border_) {
    area.x += r.border_->left;
    area.y += r.border_->top;
    area.w -= r.border_->left + r.border_->right;
    area.h -= r.border_->top  + r.border_->bottom;
  }
  return area;
}

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

struct TopCh {
  decltype(GadgetLayout::layer) layer;
  Gadget* ch;
  static bool comp(TopCh const& a, TopCh const& b) {return a.layer < b.layer;}
};


Gadget* GBox::doPick(GPoint const& pt, GPickMode& pick) {  bool specialclip = modes_.unclippedLayers && layout_.hasLayers;
  if (!specialclip && !Gadget::doPick(pt, pick)) return nullptr;
  Gadget* found{};

  // the box contains gadgets with a layer > 0 that must be treated first
  if (layout_.hasLayers) {
    //TopCh topch[children_.size()];  // undef on windows
    std::vector<TopCh> topch(children_.size());
    size_t topcount = 0;

    for (auto& ch : children_) {
      if (ch->layout_.layer > 0) {
        topch[topcount++] = {ch->layout_.layer, ch};
      }
    }

    if (topcount == 0) {
      layout_.hasLayers = 0;  // needed elsewhere
    }
    // optim to avoid sorting if only one child
    else if (topcount == 1) {
      if (auto* g = topch[0].ch->doPick(pt, pick)) {
        found = g;
        goto END;
      }
    }
    else {
      // stable_sort preserves order
      //std::stable_sort(topch, topch+topcount, TopCh::comp);
      std::stable_sort(topch.begin(), topch.begin() + topcount, TopCh::comp);

      // pick children in reverse order (note: => ssize_t not size_t!)
      for (ssize_t k = topcount-1; k >= 0; --k) {
        if (auto* g = topch[k].ch->doPick(pt, pick)) {
          found = g;
          goto END;
        }
      }
    }
  }

  if (specialclip && !Gadget::doPick(pt, pick)) return nullptr;

  // pick borders
  if (auto* blist = bchildren_) {
    for (auto& it : *blist) {
      if (it.child_->isShown() && !it.child_->isTopLevel()) {
        if (auto* g = it.child_->doPick(pt, pick)) {
          found = g;
          goto END;
        }
      }
    }
  }
  // pick children in reverse order: from the end.
  for (auto ch = children_.rbegin(); ch != children_.rend(); ++ch) {
    if ((*ch)->isShown() && !(*ch)->isTopLevel() && (*ch)->layout_.layer == 0) {
      if (auto* g = (*ch)->doPick(pt, pick)) {
        found = g;
        goto END;
      }
    }
  }

  END:
  if (modes_.catchEvents && !pick.find_) {
    pick.catchedevent_ = true;
    pick.catchedchild_ = found;
    return this;
  }

  return found ? found : this;
}

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

GDim GBox::computeInsideSize(GRender& r) {
  GDim size;
  if (children_.empty()) return size;

  switch (modes_.dir) {
    case
      GFlow::Row:
      computeInsideSizeRow(size, r);
      break;
    case GFlow::Column:
      computeInsideSizeColumn(size, r);
      break;
    case GFlow::Stack:
      computeInsideSizeStack(size, r);
      break;
    default:
      computeInsideSizeRow(size, r);
      break;
  }

  if (auto* blist = bchildren_) {
    for (auto& it : *blist) {
      it.child_->computeSize(&r, true);

      switch (it.place_) {
        case Place::Top:
        case Place::ToolBar:
        case Place::Bottom:
          size.h += it.child_->h();
          break;

        case Place::Left:
        case Place::Right:
          size.w += it.child_->w();
          break;

        default:
          break;
      }
    }
  }

  // align text in menus, etc.
  //if (modes_.leftAlign)  size.w += GPrefs::iconWidth();
  //if (modes_.rightAlign) size.w += GPrefs::menuTab();
  if (modes_.menuLayout) {
    size.w += GPrefs::iconWidth() + GPrefs::menuTab();
  }
  return size;
}


void GBox::computeInsideSizeStack(GDim& boxsize, GRender& r) {
  float ch_w{}, ch_h{};

  for (auto& ch : children_) {  // childlist pas children() !
    if (ch->isShown()) {
      ch->computeSize(&r, true);
      if (!ch->layout_.activePos && !ch->isTopLevel()) {
        ch_w = std::max(ch_w, ch->w());
        ch_h = std::max(ch_h, ch->h());
      }
    }
  }
  /* cerr << "computeInsideSizeStack "
  << " " << ch_w
  << " " << ch_h
  << endl;*/
  boxsize.w = ch_w;
  boxsize.h = ch_h;
}


void GBox::computeInsideSizeColumn(GDim& boxsize, GRender& r) {
  bool one_found = false;
  float ch_w{};

  for (auto& ch : children_) {  // childlist pas children() !
    if (ch->isShown()) {
      ch->computeSize(&r, true);
      if (!ch->layout_.activePos && !ch->isTopLevel()) {
        ch_w = std::max(ch_w, ch->w());
        boxsize.h += ch->h() + r.gap_.mainspacing;
        one_found = true;
      }
    }
  }
  boxsize.w += ch_w;
  if (one_found) boxsize.h -= r.gap_.mainspacing;
}


void GBox::computeInsideSizeRow(GDim& boxsize, GRender& r) {
  float ch_w{}, ch_h{};
  float spacing = r.gap_.mainspacing;
  float spacing2 = r.gap_.crossspacing;
  int cols = 0, lines = 1;

  for (auto& ch : children_) {  // childlist pas children() !
    if (ch->isShown()) {
      ch->computeSize(&r, true);
      if (!ch->layout_.activePos && !ch->isTopLevel()) {
        if (isGrid(r.flow_)) {
          ch_w = std::max(ch_w, ch->w());
          ch_h = std::max(ch_h, ch->h());
          ++cols;
          if (cols >= r.flow_.gridsize) {
            cols = 0;
            lines++;
          }
        }
        else {
          ch_w += ch->frame_.w + spacing;
          ch_h = std::max(ch_h, ch->h());
          ++cols;
        }
      }
    }
  }

  if (isGrid(r.flow_)) {
    if (cols == 0) lines--;
    if (lines > 0) {
      if (lines == 1) boxsize.w += cols * (ch_w + spacing) - spacing;
      else boxsize.w += r.flow_.gridsize * (ch_w + spacing) - spacing;
      boxsize.h += lines * (ch_h + spacing2) - spacing;  // spacing2 !!
    }
  }
  else {
    if (cols > 0) ch_w -= spacing;
    boxsize.w += ch_w;
    boxsize.h += ch_h;
  }
}

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

void GBox::pack(GRender* parr) {
  GRender r(*this, parr);
  auto ch_area = childrenArea(r);
  packBorders(ch_area, r);  // changes ch_area !
  if (children_.empty()) return;

  switch (modes_.dir) {
    case GFlow::Row: packRow(ch_area, r); break;
    case GFlow::Column: packColumn(ch_area, r); break;
    case GFlow::Pie: packPie(ch_area, r); break;
    case GFlow::Stack: packStack(ch_area, r); break;
    default: packRow(ch_area, r); break;
  }
}

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

void GBox::packStack(GRect& ch_area, GRender& r) {
  float wmax{}, hmax{};

  for (auto& ch : children_) {
    if (ch->isShown() && !ch->isTopLevel()) {
      if (ch->layout_.activePos) {
        // NB: pos gadgets dont change the box size and may thus be clipped
        ch->packPos(this->wx(), this->wy());
      }
      else {
        if (ch->w() > wmax) wmax = ch->w();
        if (ch->h() > hmax) hmax = ch->h();
      }
    }
  }

  if (w() > wmax) wmax = w();
  if (h() > hmax) hmax = h();

  for (auto& ch : children_) {
    if (ch->isShown() && !ch->isTopLevel()) {
      GRect& chframe = ch->frame_;
      GRect chframe_init = chframe;

      if (ch->layout_.activePos) {
        ch->pack(&r);
        ch->movedResizedCB(chframe_init);
        continue;
      }
      /*
      cerr << "packstack "<<ch_area.x
      << " " << ch_area.y
      << " " << wmax
      << " " << hmax
      << endl;*/
      chframe = {ch_area.x, ch_area.y, wmax, hmax};
      ch->pack(&r);
      ch->movedResizedCB(chframe_init);
    }
  }
}

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

void GBox::packColumn(GRect& ch_area, GRender& r) {
  auto spacing = r.gap_.mainspacing;
  int count{}, flex_count{};
  float used{}; //, maxval{};

  for (auto& ch : children_) {
    if (ch->isShown() && !ch->isTopLevel()) {
      if (ch->layout_.activePos) {
        // NB: Pos gadgets dont change the box size and may thus be clipped
        ch->packPos(frame_.x, frame_.y);
      }
      else {
        ++count;
        flex_count += ch->columnFlex(this);
        used += ch->frame_.h + spacing;
      }
    }
  }

  if (count > 0) used -= spacing;
  float remain_h = ch_area.h - used;
  float ch_y = 0;

  // remain is set to 0 except if box align = Extend on y axis
  switch (r.align_.main) {
    case GAlign::Extend:
      ch_y = ch_area.y;
      if (flex_count > 0) remain_h = remain_h / flex_count;
      break;
    case GAlign::Start:
      ch_y = ch_area.y;
      remain_h = 0.f;
      break;
    case GAlign::End:
      ch_y = ch_area.y + remain_h;
      remain_h = 0.f;
      break;
    case GAlign::Center:
      ch_y = ch_area.y + remain_h / 2.f;
      remain_h = 0.f;
      break;
    case GAlign::Equal:
      ch_y = ch_area.y;
      remain_h = 0.f;
      break;
  }

  for (auto& ch : children_) {
    if (ch->isShown() && !ch->isTopLevel()) {
      GRect& chframe = ch->frame_;
      GRect chframe_init = chframe;

      if (ch->layout_.activePos) {
        ch->pack(&r);
        ch->movedResizedCB(chframe_init);
        continue;
      }
      /*
      if (ch->layout_.wpercent > 0)
        chframe.w = ch_area.w * ch->layout_.wpercent / 100.f;

      if (ch->layout_.hpercent > 0)
        chframe.h = ch_area.h * ch->layout_.hpercent / 100.f;
       */
      if (r.size_.wunit == GSize::Percent) {
        chframe.w = ch_area.w * r.size_.w / 100.f;
      }
      if (r.size_.hunit == GSize::Percent) {
        chframe.h = ch_area.h * r.size_.h / 100.f;
      }
      else if (r.align_.main == GAlign::Equal) {
        // all children have same h
        chframe.h = (ch_area.h - spacing * (count-1)) / count;
      }
      else {
        // adjust child vertically if it is Flex (and if remain is not 0)
        chframe.h += ch->columnFlex(this) * remain_h;
      }

      switch (r.align_.cross) {
        case GAlign::Extend:
        case GAlign::Equal:
          // adjust child horizontally except if fixed (Flex == 0)
          if (ch->rowFlex(this) == 0) {              // fixed, dont-resize
            chframe = {                                 // centered on x axis
              ch_area.x + (ch_area.w-chframe.w)/2.f,
              ch_y, chframe.w, chframe.h
            };
          }
          else chframe = {ch_area.x, ch_y, ch_area.w, chframe.h};
          break;
        case GAlign::Start:
          chframe = {ch_area.x, ch_y, chframe.w, chframe.h};
          break;
        case GAlign::End:
          chframe = {ch_area.x + ch_area.w-chframe.w, ch_y, chframe.w, chframe.h};
          break;
        case GAlign::Center:
          chframe = {ch_area.x + (ch_area.w-chframe.w)/2.f, ch_y, chframe.w, chframe.h};
          break;
      }

      ch->pack(&r);
      // arrondir a entier superieur pour eviter que ca "bave"
      ch_y += int(chframe.h + spacing + 0.999);
      ch->movedResizedCB(chframe_init);
    }
  }
}

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
// NB: prend en compte Grid contrairement a packChildrenColumn()

void GBox::packRow(GRect& ch_area, GRender& r) {
  auto spacing = r.gap_.mainspacing;
  auto spacing2 = r.gap_.crossspacing;
  int count{}, flex_count{}, lines{1}, grid_cols{};
  float used{}, grid_w{}, grid_h{};  //  maxval{},

  for (auto& ch : children_) {  // childlist pas children() !
    if (ch->isShown() && !ch->isTopLevel()) {
      if (ch->layout_.activePos) {
        // NB: Pos gadgets dont change the box size and may thus be clipped
        ch->packPos(frame_.x, frame_.y);
      }
      else {
        if (isGrid(r.flow_)) {
          grid_w = std::max(grid_w, ch->frame_.w);
          grid_h = std::max(grid_h, ch->frame_.h);
          ++grid_cols;
          if (grid_cols >= r.flow_.gridsize) {
            grid_cols = 0;
            lines++;
          }
        }
        else {
          used += ch->frame_.w + spacing;
          count++;
          flex_count += ch->rowFlex(this);
        }
      }
    }
  } // endfor

  if (isGrid(r.flow_)) {
    if (grid_cols == 0) lines--;
    if (lines == 0) used = 0;
    else if (lines == 1) used = grid_cols * (grid_w + spacing) - spacing;
    else used = r.flow_.gridsize * (grid_w + spacing) - spacing;
  }
  else {
    if (count > 0) used -= spacing;
  }

  float remain_w = ch_area.w - used;
  float ch_x = 0;

  // remain is set to 0 except if the child is expanded
  switch (r.align_.main) {
    case GAlign::Extend:
      ch_x = ch_area.x;
      if (flex_count > 0) remain_w /= flex_count;  // pourrait etre en %
      break;
    case GAlign::Start:
      ch_x = ch_area.x;
      remain_w = 0.f;
      break;
    case GAlign::End:
      ch_x = ch_area.x + remain_w;
      remain_w = 0.f;
      break;
    case GAlign::Center:
      ch_x = ch_area.x + remain_w / 2.f;
      remain_w = 0.f;
      break;
    case GAlign::Equal:
      ch_x = ch_area.x;
      remain_w = 0.f;
      break;
  }

  float ch_x0 = ch_x;
  float ch_y = ch_area.y;
  float remain_h = 0;
  grid_cols = 0;

  for (auto& ch : children_) {
    if (ch->isShown() && !ch->isTopLevel()) {
      GRect& chframe = ch->frame_;
      GRect chframe_init = chframe;

      if (ch->layout_.activePos) {
        ch->pack(&r);
        ch->movedResizedCB(chframe_init);
        continue;
      }

      if (isGrid(r.flow_)) {
        chframe.w = grid_w;
        chframe.h = grid_h;
        remain_w = ch_area.h - (grid_h + spacing2) * lines - spacing2; //spacing2!!
      }
      else {
        /*
        if (ch->layout_.hpercent > 0) {
          chframe.h = ch_area.h * ch->layout_.hpercent / 100.f;
        }
        if (ch->layout_.wpercent > 0) {
          chframe.w = ch_area.w * ch->layout_.wpercent / 100.f;
        }
         */
        if (r.size_.hunit == GSize::Percent) {
          chframe.h = ch_area.h * r.size_.h / 100.f;
        }

        if (r.size_.wunit == GSize::Percent) {
          chframe.w = ch_area.w * r.size_.w / 100.f;
        }
        else if (r.align_.main == GAlign::Equal) {
          // all children have same w
          chframe.w = (ch_area.w - spacing * (count-1)) / count;
        }
        else {
          // adjust child horizontally if it is Flex (and if remain is not 0)
          chframe.w += ch->rowFlex(this) * remain_w;
        }
        remain_h = ch_area.h - chframe.h;
      }

      switch (r.align_.cross) {
        case GAlign::Extend:
        case GAlign::Equal:
          // adjust child vertically except if Fixed or grid
          if (ch->columnFlex(this) == 0 || isGrid(r.flow_)) { // fixed
            chframe = {
              ch_x, ch_y + remain_h/2.f,
              chframe.w, chframe.h
            };
          }
          else chframe = {ch_x, ch_y, chframe.w, ch_area.h}; // ch_area.h !
          break;
        case GAlign::Start:
          chframe = {ch_x, ch_y, chframe.w, chframe.h};
          break;
        case GAlign::End:
          chframe = {ch_x, ch_y + remain_h, chframe.w, chframe.h};
          break;
        case GAlign::Center:
          chframe = {ch_x, ch_y + remain_h/2.f, chframe.w, chframe.h};
          break;
      }
      ch->pack(&r);
      // arrondir a entier superieur pour eviter que ca "bave"
      ch_x += int(chframe.w + spacing + 0.999);

      if (isGrid(r.flow_)) {
        ++grid_cols;
        if (grid_cols >= r.flow_.gridsize) {
          grid_cols = 0;
          ch_x = ch_x0;
          ch_y += grid_h + spacing2;  //spacing2
        }
      }
      ch->movedResizedCB(chframe_init);
    }
  }
}

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

void GBox::packPie(GRect& ch_area, GRender& r) {
  //auto spacing = r.gap_->spacing;
  int count{}, flex_count{};
  float totalval{};

  for (auto& ch : children_) {
    if (ch->isShown() && !ch->isTopLevel()) {
      if (ch->layout_.activePos) {
        // NB: pos gadgets dont change the box size and may thus be clipped
        ch->packPos(this->wx(), this->wy());
      }
      else {
        ++count;
        flex_count += ch->columnFlex(this);
        if (auto* ci = ch->toValueItem()) totalval += ci->value();
      }
    }
  }

  float start{};

  for (auto& ch : children_) {
    if (ch->isShown() && !ch->isTopLevel()) {
      GRect chframe_init = ch->frame_;

      if (ch->layout_.activePos) {
        ch->pack(&r);
        ch->movedResizedCB(chframe_init);
        continue;
      }
      float arc{};
      if (auto* ci = ch->toValueItem()) {
        arc = ci->value() / totalval;
        ci->offset_ = arc;
        ci->modes_.dir = GFlow::Pie;
      }
      else {
        arc = 1.f / count;
      }
      ch->setAngleStart(start * 360);
      ch->setAngleExtent(arc * 360);
      start += arc;
      ch->frame_ = {ch_area.x, ch_area.y, ch_area.w, ch_area.h};
      ch->movedResizedCB(chframe_init);
    }
  }
}

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

void GBox::packBorders(GRect& charea, GRender& r) {
  auto* blist = bchildren_;
  if (!blist) return;

  const GRect& boxframe = this->frame_;
  float top1{}, top2{}, bottom{}, left{}, right{};

  for (auto& it : *blist) {
    switch (it.place_) {
      case Place::Top:
        top1 = it.child_->frame_.h;
        break;
      case Place::ToolBar:
        top2 = it.child_->frame_.h;
        break;
      case Place::Bottom:
        bottom = it.child_->frame_.h;
        break;
      case Place::Left:
        left = it.child_->frame_.w;
        break;
      case Place::Right:
        right = it.child_->frame_.w;
        break;
      default:
        break;
    }
  }

  for (auto& it : *blist) {
    auto& b = it.child_;
    GRect chframe_init = b->frame_;

    switch (it.place_) {
      case Place::Top: {
        b->frame_ = {boxframe.x, boxframe.y, boxframe.w, top1};
        b->pack(&r);
        charea.y += top1;
        charea.h -= top1;
        b->movedResizedCB(chframe_init);
      } break;

      case Place::ToolBar: {
        b->frame_ = {boxframe.x, boxframe.y + top1, boxframe.w, top2};
        b->pack(&r);
        charea.y += top2;
        charea.h -= top2;
        b->movedResizedCB(chframe_init);
      } break;

      case Place::Bottom: {
        b->frame_ = {
          boxframe.x, boxframe.y + boxframe.h - bottom,
          boxframe.w - right, bottom
        };
        b->pack(&r);
        charea.h -= bottom;
        b->movedResizedCB(chframe_init);
      } break;

      case Place::Right: {
        b->frame_ = {
          boxframe.x + boxframe.w - right, boxframe.y + top1+top2,
          right, boxframe.h - top1-top2 - bottom
        };
        b->pack(&r);
        charea.w -= right;
        b->movedResizedCB(chframe_init);
      } break;

      case Place::Left: {
        b->frame_ = {
          boxframe.x, boxframe.y + top1+top2,
          left, boxframe.h - top1-top2 - bottom
        };
        b->pack(&r);
        charea.x += left;
        charea.w -= left;
        b->movedResizedCB(chframe_init);
      } break;

      default:
        break;
    }
  }
}

}
