//
//  Sliders & Scrollbars 
//
// Guit GUI Toolkit (https://www.telecom-paris.fr/~elc/guit/)
// Copyright 2020 Eric Lecolinet (eric.lecolinet@gmail.com) - Telecom Paris
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//

#ifndef Guit_Slider_hpp
#define Guit_Slider_hpp
#include <gvalueitem.hpp>
namespace guit {

/// Slider gadget.
/// The slider's value depends on a GNumber prop (see operator()()). Changing the value of this prop
/// updates the slider. Conversely, moving the slider's knob changes the prop's value.
/// The default min is 0 and the default max is 100.
///
/// Gadgets can be synchronized by sharing props:
/// @code
///     auto& val = Float();
///     auto& slider = Slider() << val;
///     auto& textf  = TextField() << val;
/// @endcode
/// Both gadgets are then synced, so that changing one changes the other one.
/// Moreover, the textfield will always display a float and force the user to enter a float.
///
/// By default, a slider is horizontal. It can be made vertical as follows:
/// @code
///      myslider << GFlow::column;
/// @endcode
///
class GSlider : public GValueItem {
protected:
  GSlider(GadgetType&, GString const& css_and_value);

public:
  GSlider(GString const& css_and_value = "");
  GSlider(float value);
  GSlider& clone(GClone const&) override;

  /// changes the minimum and the maximum value.
  /// The default min is 0 and the default max is 100.
  /// @see GValueItem::setRange()
  /// @{
  GSlider& setRange(float min, float max);

  using GType = GadgetType_<GSlider,GValueItem>;
  static GType& Type();
  GType& type() const override {return Type();}

  /// @internal
  /// @{
  GFlow::Kind getFlow() override;
  void paintContent(GRenderGraphics&, GRender&, GFlow::Kind) override;
  /// @}
};

/// Creates a GSlider.
/// @{
GSlider& Slider(GString const& css_and_value = "");
GSlider& Slider(float value);
/// @}

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

/// Scroll bar.
/// @see GSlider.
class GScrollBar : public GSlider {
protected:
  GScrollBar(GadgetType&, GString const& css_and_value);

public:
  GScrollBar(GString const& css_and_value = "");
  GScrollBar& clone(GClone const&) override;

  Place place() const override {return Place::ScrollBar;}

  using GType = GadgetType_<GScrollBar,GSlider>;
  static GType& Type();
  GType& type() const override {return Type();}

  /// @internal
  /// @{
  void paintContent(GRenderGraphics&, GRender&, GFlow::Kind) override;
  void mouseDrag(GMouseEvent&) override;
  /// @}
};

/// Creates a GScrollBar.
GScrollBar& ScrollBar(GString const& css_and_value = "");

}
#endif
