//
// Files and file paths
//
// Guit GUI Toolkit (https://www.telecom-paris.fr/~elc/guit/)
// Copyright 2020 Eric Lecolinet (eric.lecolinet@gmail.com) - Telecom Paris
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//

#ifndef Guit_File_hpp
#define Guit_File_hpp
#include <fstream>
#include <gstring.hpp>
namespace guit {

/// Opens a stream for reading, taking into account magic characters.
/// - If _path_ starts with / or \ or .  just opens it
/// - If _path_ starts with ~ or : or ! this symbol is replaced by calling normPath()
/// - Otherwise, tries to open _path_, then, if not found,  opens the first existing dir/path file
///  by considering  the _dir_ elements of GApp::resourceDirs() in a sequence
bool openFile(std::ifstream&, GString const& path, const char* locale = nullptr);

/// Opens a stream for writing, taking into account magic characters.
/// - If _path_ starts with ~ or : or ! the first symbol is replaced by calling normPath()
/// - Otherwise, opens _path_
bool openFile(std::ofstream&, GString const& path, const char* locale = nullptr);


/// Creates a file for reading and calls OpenFile().
class GIFile : public std::ifstream  {
public:
  GIFile(GString const& path, const char* locale = {}) {
    openFile(*this, path, locale);
  }
};

/// Creates a file for writing and calls OpenFile().
class GOFile : public std::ofstream {
public:
  GOFile(GString const& path, const char* locale = {}) {
    openFile(*this, path, locale);
  }
};

/// Expands ~ : ! magic characters at the beginning of filenames.
/// If _filename_ starts with ~ or : or !  this symbol is replaced as follows:
/// -  ~ is replaced by GApp::homeDir()
/// -  : is replaced by GApp::resourceDir()
/// -  ! is replaced by GApp::appDir()
/// Moreoever, if _add_slash_ is true a final slash is added to filename if it does not have one.
GString normPath(GString const& path, bool add_slash = false);

/// Returns true if this file exists.
/// Resolves ~ and : if the path starts with these characters (see fileexpand()).
bool fileExist(GString const& path);

/// copies from_file into to_file.
/// creates or overwrites to_file if needed.
bool copyFile(GString const& from_file, GString const& to_file);

/// Calls the C system() function.
int system(GString const& command);

/// Creates a directory.
bool makeDir(GString const& path);

/// Returns the parent directory.
GString parentPath(GString const& path);

/// Returns the filename component.
/// Keeps filename extension if _with_extension_ is true.
GString basename(GString const& path, bool with_extension = true);

/// Returns the directory component.
GString dirname(GString const& path);

/// Returns the file extension.
/// Keeps the dot before the extension if _with_dot_ is true.
GString extname(GString const& path, bool with_dot = true);

}
#endif
