//
//  Multitouch detection and Cursor management
//
// Guit GUI Toolkit (https://www.telecom-paris.fr/~elc/guit/)
// Copyright 2020 Eric Lecolinet (eric.lecolinet@gmail.com) - Telecom Paris
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//

#ifdef __APPLE__

#ifndef GuitTouch_h
#define GuitTouch_h
#include <gcursor.hpp>
namespace guit {

class GTouch {
public:
  enum Phase {
    NotTracking = 0,
    Start       = 1,
    Hover       = 2,
    BeginTouch  = 3,
    Touching    = 4,
    EndTouch    = 5,
    Linger      = 6,
    OutOfRange  = 7
  };

  int32_t frame;      // current frame
  double time;        // event timestamp
  int32_t touchid;    // unique for life of touch per device
  int32_t phase;      // phase (distance from touchpad)
  int32_t fingerid;
  int32_t handid;
  guit::GPoint pos;
  guit::GPoint speed;
  float area;         // touch area (pseudo-pressure), between 0,0 and 1,0
  int32_t aa;
  float angle;        // angle of the touch ellipse
  float majoraxis;    // major axis of the touch ellipse
  float minoraxis;    // minor axis of the touch ellipse
  guit::GPoint abspos;
  guit::GPoint absspeed;
  int32_t bb;
  int32_t cc;
  float density;
};

struct GTouchDevice;

class GTouchCursor {
public:
  bool isNative() const {return cursor_->nativecursor_;}
  void touchDown(GTouch const&);
  void touchUp(GTouch const&);
  void touchMove(GTouch const&);
  void mouseDown();
  void mouseUp();
  void sendButtonEvent(bool press);
  void sendMoveEvent();

private:
  friend class GTouchManager;
  friend class GTouchPad;
  int32_t touchid_{};
  GPoint mouse_, lastmouse_, locmouse_;
  GPoint lastpos_;
  double lasttime_{};
  GCursor* cursor_{};
};


class GTouchPad {
public:
  GTouchCursor& createCursor2(GColor& cursor_color);
  GTouchCursor* cursor1() {return curs1_;}
  GTouchCursor* cursor2() {return curs2_;}

protected:
  friend class GTouchManager;
  GTouchPad(GTouchDevice*, bool nativecursor);
  void touchCB(GTouch const* touches, int touchCount, double timestamp);
  GTouchDevice *device_{};
  GTouchCursor *curs1_{}, *curs2_{};
};


class GTouchManager {
public:
  /// returns the touchmanager, craeting it if needed.
  static GTouchManager& instance();

  /// Starts the touch manager.
  /// returns the number of touchpads, 0 if none found.
  size_t init(GWindow& window);
  void start();
  void stop();

  float mouseGain() const {return mousegain_;}
  void  setMouseGain(float val) {mousegain_ = val;}

  double timeResolution() const {return timeres_;}
  void   setTimeResolution(double val) {timeres_ = val;}

  //bool captureMouse() const {return capturemouse_;}
  //void setCaptureMouse(bool val) {capturemouse_ = val;}

  //float spaceResolution() const {return spaceres_;}
  //void  setSpaceResolution(float val) {spaceres_ = val;}

  GTouchPad* getPad(unsigned int channel);
  size_t padCount() const {return pads_.size();}
  
protected:
  friend class GTouchPad;
  friend class GTouchCursor;

  GTouchManager();
  GTouchManager(GTouchManager const&) = delete;
  GTouchManager(GTouchManager const&&) = delete;
  GTouchManager& operator=(GTouchManager const&) = delete;
  GTouchManager& operator=(GTouchManager const&&) = delete;

  GTouchCursor* addCursor(GColor&);
  static void focusInWindowCB(GEvent&);
  static void focusOutWindowCB(GEvent&);
  static bool mouseButtonCB(GMouseEvent&);
  static bool mouseMotionCB(GMouseEvent&);
  static bool wheelCB(GMouseEvent&);
  static int touchCB(GTouchDevice* d, const GTouch* touches, int touchcount,
                     double timestamp, int frame);
  GDim screensize_;
  bool running_{};
  float mousegain_{};
  double timeres_{};
  std::vector<GTouchPad*> pads_;
  std::vector<GTouchCursor*> cursors_;
  GTouchCursor* nativecurs_{};
  GWindow* win_{};
};

}
#endif

#endif
