//
//  Multitouch detection and Cursor management
//
// Guit GUI Toolkit (https://www.telecom-paris.fr/~elc/guit/)
// Copyright 2020 Eric Lecolinet (eric.lecolinet@gmail.com) - Telecom Paris
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//

#ifdef __APPLE__

#include <iostream>
#include <cstdlib>
#include <Foundation/Foundation.h>
#include <SDL.h>
#include <guit.hpp>
#include <gcursor.hpp>
#include <gcontrol.hpp>
#include <gmactouch.hpp>
using namespace std;
using namespace guit;

extern "C" {
// returns a pointer to the default device (the trackpad?)
GTouchDevice* MTDeviceCreateDefault();

// returns a CFMutableArrayRef array of all multitouch devices
NSMutableArray* MTDeviceCreateList();

// Touch callback function
typedef int (*MTContactFrameCallback)(GTouchDevice*, GTouch const* contacts,
                                      int numContacts, double timestamp, int frame);

// registers a device's frame callback to your callback function
void MTRegisterContactFrameCallback(GTouchDevice*, MTContactFrameCallback);
void MTUnregisterContactFrameCallback(GTouchDevice*, MTContactFrameCallback);

// start sending events
void MTDeviceStart(GTouchDevice*, int32_t = 0);
void MTDeviceStop(GTouchDevice*);
void MTDeviceRelease(GTouchDevice*);

bool MTDeviceIsAvailable();
bool MTDeviceIsRunning(GTouchDevice*);
double MTAbsoluteTimeGetCurrent();

int MTDeviceGetSensorDimensions(GTouchDevice*, int32_t *rows, int32_t *cols);
int MTDeviceGetSensorSurfaceDimensions(GTouchDevice*, int32_t *width, int32_t *height);

bool MTDeviceIsValid(GTouchDevice*);
bool MTDeviceIsOpaqueSurface(GTouchDevice*);
OSStatus MTDeviceGetFamilyID(GTouchDevice*, int32_t* familyId);
OSStatus MTDeviceGetDeviceID(GTouchDevice*, uint64_t* ID) __attribute__ ((weak_import));
OSStatus MTDeviceGetDriverType(GTouchDevice*, int32_t*);
OSStatus MTDeviceGetGUID(GTouchDevice*, uuid_t* guid);
}

namespace guit {

static GTouchManager* inst_{};

GTouchManager& GTouchManager::instance() {
  if (!inst_) inst_ = new GTouchManager();
  return *inst_;
}

GTouchManager::GTouchManager() {
  mousegain_ = 1.f;
  timeres_ = 0.02;
  //spaceres_ = 0.002f;
  //outsidemargin_ = 5.f;
}

GTouchPad* GTouchManager::getPad(unsigned int chan) {
  if (chan < pads_.size()) return pads_[chan];
  else return nullptr;
}

size_t GTouchManager::init(GWindow& win) {
  screensize_ = GCursor::screenSize();
  win_ = &win;
  
  NSMutableArray* devlist = MTDeviceCreateList();
  if (!devlist) {
    gerror("GTouchManager::init","could not retrieve touch device");
    return 0;
  }
  // retrieve touchpad size and register callback for retrieving touch events
  bool first = true;
  for (size_t k = 0; k < [devlist count]; k++) {
    if (auto* d = (__bridge GTouchDevice*)[devlist objectAtIndex:k]) {
      //int w = 0, h = 0;
      //MTDeviceGetSensorSurfaceDimensions(d, &w, &h); // get trackpad size

      auto* pad = new GTouchPad(d, first);
      first = false;

      pads_.push_back(pad);
      // register callback that gets touch events
      MTRegisterContactFrameCallback(d, GTouchManager::touchCB);
    }
  }

  win
  << on::focus / focusInWindowCB
  << on::focusout / focusOutWindowCB;
  return pads_.size();
}


void GTouchManager::start() {
  if (running_) return;
  running_ = true;

  // set bevahior of channel 0, the standard mouse event channel
  GEvent::setMouseButtonHandler(0, mouseButtonCB);
  GEvent::setMouseMotionHandler(0, mouseMotionCB);
  GEvent::setWheelHandler(0, wheelCB);

  SDL_EventState(SDL_FINGERDOWN, SDL_IGNORE);
  SDL_EventState(SDL_FINGERUP, SDL_IGNORE);
  SDL_EventState(SDL_FINGERMOTION, SDL_IGNORE);
  SDL_EventState(SDL_MULTIGESTURE, SDL_IGNORE);
  SDL_EventState(SDL_DOLLARGESTURE, SDL_IGNORE);
  SDL_EventState(SDL_MOUSEMOTION, SDL_IGNORE);
  SDL_EventState(SDL_MOUSEWHEEL, SDL_IGNORE);

  for (auto& p : pads_) MTDeviceStart(p->device_);
  for (auto& c : cursors_) c->touchid_ = 0;
}


void GTouchManager::stop() {
  running_ = false;

  GCursor::cursor(0)->show(true);
  GCursor::cursor(0)->freeze(false);

  GEvent::setMouseButtonHandler(0, nullptr);
  GEvent::setMouseMotionHandler(0, nullptr);
  GEvent::setWheelHandler(0, nullptr);

  SDL_EventState(SDL_FINGERDOWN, SDL_ENABLE);
  SDL_EventState(SDL_FINGERUP, SDL_ENABLE);
  SDL_EventState(SDL_FINGERMOTION, SDL_ENABLE);
  SDL_EventState(SDL_MULTIGESTURE, SDL_ENABLE);
  SDL_EventState(SDL_DOLLARGESTURE, SDL_ENABLE);
  SDL_EventState(SDL_MOUSEMOTION, SDL_ENABLE);
  SDL_EventState(SDL_MOUSEWHEEL, SDL_ENABLE);

  for (auto& p : pads_) MTDeviceStop(p->device_);
  for (auto& c : cursors_) c->touchid_ = 0;
}


void GTouchManager::focusInWindowCB(GEvent& e) {
  if (!inst_->running_) return;
  //cerr << "focusin " << e.window() << endl;
  //GCursor::cursor(0)->freeze(true);
  for (auto& p : inst_->pads_) MTDeviceStart(p->device_);
  for (auto& c : inst_->cursors_) c->touchid_ = 0;
}

void GTouchManager::focusOutWindowCB(GEvent& e) {
  if (!inst_->running_) return;
  //cerr << "focusout " << e.window() << endl;
  GCursor::cursor(0)->freeze(false);
  for (auto& p : inst_->pads_) MTDeviceStop(p->device_);
  for (auto& c : inst_->cursors_) c->touchid_ = 0;
}

/*
void GTouchManager::enterWindowCB(GEvent& e) {
  inst_->enteredwin_ = e.window();
}
void GTouchManager::leaveWindowCB(GEvent& e) {
}*/

bool GTouchManager::wheelCB(GMouseEvent& e) {
  // cerr << "  wheelCB " << e.window()
  // << " main " << int(e.window() != inst_->win_) << endl;
  // les autres fenetres fonctionnent normalement
  return (e.window() != inst_->win_);
}

bool GTouchManager::mouseMotionCB(GMouseEvent& e) {
  // cerr << "  mouseMotionCB " << e.window()
  // << " main " << int(e.window() != inst_->win_) << endl;
  // les autres fenetres fonctionnent normalement
  return (e.window() != inst_->win_);
}

bool GTouchManager::mouseButtonCB(GMouseEvent& e) {
  /*
   cerr << "  mouseButtonCB " << e.window()
   << " main " << int(e.window() != inst_->win_)
   << " chan " << e.channel()
   << endl;
   */
  if (e.window() != inst_->win_) return true;  // process event

  // 0 = standard mouse channel, do nothing if other channel.
  if (e.channel() == 0) {
    if (e.trigger() == on::mousedown) {
      if (inst_->nativecurs_) inst_->nativecurs_->mouseDown();
    }
    else if (e.trigger() == on::mouseup) {
      if (inst_->nativecurs_) inst_->nativecurs_->mouseUp();
    }
  }
  return false;
}


// callback for trackpad events.
int GTouchManager::touchCB(GTouchDevice* d, const GTouch* touches, int touchcount,
                           double timestamp, int frame) {
  if (touchcount <= 0 || !touches) return 0;
  for (auto& p : inst_->pads_) {
    // may be called multiple times as several pads can have the same device ??
    if (p->device_ == d) p->touchCB(touches, touchcount, timestamp);
  }
  return 0;
}

GTouchCursor* GTouchManager::addCursor(GColor& color) {
  auto* c = new GTouchCursor();
  c->cursor_ = &GCursor::obtainCursor(int(inst_->cursors_.size() + 1));
  c->cursor_->setColor(color);
  c->cursor_->show(true);
  cursors_.push_back(c);
  return c;
}

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

GTouchPad::GTouchPad(GTouchDevice* d, bool nativecursor)
: device_(d) {
  curs1_ = GTouchManager::instance().addCursor(GColor::red);
  curs1_->locmouse_ = {inst_->win_->w() / 2, inst_->win_->h() / 2};
  //curs1_->touchframe_ = {0.f, 0.f, 1.f, 1.f};
  if (nativecursor) {
    inst_->nativecurs_ = curs1_;
    curs1_->cursor_->nativecursor_ = true;       // !!!!
  }
}

GTouchCursor& GTouchPad::createCursor2(GColor& color) {
  curs2_ = GTouchManager::instance().addCursor(color);
  curs2_->locmouse_ = {inst_->win_->w() / 2, inst_->win_->h() / 2};
  return *curs2_;
}

/*
void GTouchPad::split(bool main_on_right, float main_percent) {
  if (main_on_right) {
    curs1_->touchframe_ = {1.f-main_percent, 0.f, main_percent, 1.f};
    curs2_ = GTouchManager::instance().addCursor(GColor::blue);
    curs2_->touchframe_ = {0.f, 0.f, 1.f-main_percent, 1.f};
  }
  else {
    curs1_->touchframe_ = {0, 0.f, main_percent, 1.f};
    curs2_ = GTouchManager::instance().addCursor(GColor::blue);
    curs2_->touchframe_ = {main_percent, 0.f, 1.f-main_percent, 1.f};
  }
}*/

void GTouchPad::touchCB(const GTouch* touches, int touchcount, double time)
{
  for (int k = 0; k < touchcount; ++k) {
    auto& t = touches[k];
    /*
    cout << "\nTouch: " << k << " of "<< touchcount << endl;
    cout << " @ frame="<< t.frame << " id=" << t.touchID << " state=" << t.phase
    << "\n - x=" << t.pos.x << " y=" << t.pos.y
    << "\n - a=" << t.aa << " b=" << t.bb << " c=" << t.cc
     << " / density=" << t.density
     << "\n - size=" << t.area
     // << " angle=" << t.angle
     // << " / minor axis=" << t.minorAxis << " major axis=" << t.majorAxis
     << endl;
     */

    switch (t.phase) {

      case GTouch::BeginTouch:
        if (curs1_->touchid_ == 0) {
          curs1_->touchDown(t);
        }
        else if (curs2_ && curs2_->touchid_ == 0) {
          curs2_->touchDown(t);
        }
        continue; // other touches ignored

      case GTouch::EndTouch:
        if (curs1_->touchid_ == t.touchid) {
          curs1_->touchUp(t);
        }
        else if (curs2_ && curs2_->touchid_ == t.touchid) {
          curs2_->touchUp(t);
        }
        continue;

      case GTouch::Touching:
        if (curs1_->touchid_ == t.touchid) {
          curs1_->touchMove(t);
        }
        else if (curs2_ && curs2_->touchid_ == t.touchid) {
          curs2_->touchMove(t);
        }
        continue;
    }
  }
}


void GTouchCursor::touchDown(GTouch const& t) {
  /*
  cerr << ">>> beginTouch " << cursor_->ident()
  << " nat: " << int(isNative())
  << " touch:" << t.touchid
  << " mouse " << GCursor::cursor()->pos()
  << endl;
   */
  touchid_ = t.touchid;
  if (isNative()) mouse_ = GCursor::cursor()->pos();
  else mouse_ = inst_->win_->pos() + locmouse_;
  lastmouse_ = mouse_;
  lastpos_ = t.pos;
  lasttime_ = t.time;
  sendMoveEvent();
}


void GTouchCursor::touchUp(GTouch const& t) {
  /*
  cerr << "<<< endTouch " << cursor_->ident()
  << " nat: " << int(isNative())
  << " touch:" << t.touchid << endl << endl;
   */
  touchid_ = 0;
  lastmouse_ = mouse_;
  locmouse_ = mouse_ - inst_->win_->pos();
  lastpos_ = t.pos;
  lasttime_ = t.time;
}


void GTouchCursor::touchMove(GTouch const& t) {
  if (t.time - lasttime_ < inst_->timeres_) return;

  float dx = t.pos.x - lastpos_.x;
  float dy = lastpos_.y - t.pos.y;
  if (dx == 0 && dy == 0) return;
  //if (abs(dx) < inst_->spaceres_ && abs(dy) < inst_->spaceres_) continue;

  mouse_.x += dx * inst_->screensize_.w * inst_->mousegain_;
  mouse_.y += dy * inst_->screensize_.h * inst_->mousegain_;

  if (abs(mouse_.x - lastmouse_.x) < 1.f
      && abs(mouse_.y - lastmouse_.y) < 1.f) return;
  /*
  cerr << "Move " << cursor_->id()
  << " nat: " << int(isNative())
  << " touch:" << t.touchID << endl;
   */
  sendMoveEvent();
  //inst_->win_->update();   // to display the cursor
  lastmouse_ = mouse_;
  lastpos_ = t.pos;
  lasttime_ = t.time;
}

void GTouchCursor::mouseDown() {
  if (isNative()) sendButtonEvent(true);
}

void GTouchCursor::mouseUp() {
  if (isNative()) sendButtonEvent(false);
}

void GTouchCursor::sendButtonEvent(bool press) {
  /*
  if (press) cerr << "  >> press: " << mouse_ << " " << cursor_->ident()<< endl;
  else cerr << "  << release: " << mouse_ << " " << cursor_->ident()<<endl;
*/
   SDL_Event e;
  e.type = press ? SDL_MOUSEBUTTONDOWN : SDL_MOUSEBUTTONUP;
  e.button.timestamp = GCore::elapsedTime();
  e.button.windowID = inst_->win_->windowID();
  e.button.which = cursor_->channel();
  e.button.button = SDL_BUTTON_LEFT;
  e.button.state = press ? SDL_PRESSED : SDL_RELEASED;
  e.button.clicks = 1;
  e.button.x = int(mouse_.x - inst_->win_->x());  // coordinates relative to window
  e.button.y = int(mouse_.y - inst_->win_->y());
  SDL_PushEvent(&e);
}

void GTouchCursor::sendMoveEvent() {
  //cerr << "&&motion: " << mouse.x << " " << mouse.y << " win: " << curwin << endl;
  SDL_Event e;
  e.type = SDL_MOUSEMOTION;
  e.motion.timestamp = GCore::elapsedTime();
  e.motion.windowID = inst_->win_->windowID();
  e.motion.which = cursor_->channel();
  e.motion.state = SDL_BUTTON_LMASK;
  e.motion.x = int(mouse_.x - inst_->win_->x());
  e.motion.y = int(mouse_.y - inst_->win_->y());
  e.motion.xrel = int(mouse_.x - lastmouse_.x);
  e.motion.yrel = int(mouse_.y - lastmouse_.y);
  SDL_PushEvent(&e);
}

}
#endif

/*
bool isInside(GPoint const& pos, GRect const& frame) {
  return pos.x >= frame.x && pos.y >= frame.y
  && pos.x <= frame.x+frame.w && pos.y <= frame.y+frame.h;
}
 */
