//
//  Native SDL Layer
//  partly inspired from Will Usher's tutorial: http://www.willusher.io/pages/sdl2/
//
// Guit GUI Toolkit (https://www.telecom-paris.fr/~elc/guit/)
// Copyright 2020 Eric Lecolinet (eric.lecolinet@gmail.com) - Telecom Paris
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//


#include <gnative.hpp>  // must be first!
#include <iostream>
#include <functional>
#include <SDL_timer.h>
#include <SDL_image.h>
#include <SDL_opengl.h>
#include <SDL_ttf.h>
#include <SDL_syswm.h>
#include <gfile.hpp>
#include <gapp.hpp>
#include <gtrigger.hpp>
#include <gwindow.hpp>
#include <ganim.hpp>
#include <gcursor.hpp>
#include <ggraphics.hpp>
#include <gicon.hpp>
#include <gfont.hpp>
#include <gfontbase.hpp>
#include <gcontrolimpl.hpp>
using namespace std;
namespace guit {

const GKeyCode
GEvent::EscapeKey = SDLK_ESCAPE,
GEvent::TabKey = SDLK_KP_TAB,
GEvent::SpaceKey = SDLK_SPACE,
GEvent::BackspaceKey = SDLK_BACKSPACE,
GEvent::ReturnKey = SDLK_RETURN2,
GEvent::EnterKey = SDLK_RETURN,
GEvent::DeleteKey = SDLK_DELETE,

GEvent::LeftKey = SDLK_LEFT,
GEvent::RightKey = SDLK_RIGHT,
GEvent::UpKey = SDLK_UP,
GEvent::DownKey = SDLK_DOWN,
GEvent::HomeKey = SDLK_HOME,
GEvent::EndKey = SDLK_END,

GEvent::ShiftKey = SDLK_LSHIFT,  // SDLK_RSHIFT
GEvent::CtrlKey = SDLK_LCTRL, // SDLK_RCTRL
GEvent::AltKey = SDLK_LALT,      // SDLK_RALT
GEvent::MetaKey = SDLK_LGUI,     // SDLK_RGUI
GEvent::AltGrKey = SDLK_MODE,
GEvent::CapsKey = SDLK_CAPSLOCK,
GEvent::NumKey = SDLK_NUMLOCKCLEAR;

const GCursor::Shape
GCursor::Arrow = SDL_SYSTEM_CURSOR_ARROW,
GCursor::Crosshair = SDL_SYSTEM_CURSOR_CROSSHAIR,
GCursor::IBeam = SDL_SYSTEM_CURSOR_IBEAM,
GCursor::Wait = SDL_SYSTEM_CURSOR_WAIT,
GCursor::WaitArrow = SDL_SYSTEM_CURSOR_WAITARROW,
GCursor::Resize_W_E = SDL_SYSTEM_CURSOR_SIZEWE,
GCursor::Resize_N_S = SDL_SYSTEM_CURSOR_SIZENS,
GCursor::Resize_NE_SW = SDL_SYSTEM_CURSOR_SIZENWSE,
GCursor::Resize_NW_SE = SDL_SYSTEM_CURSOR_SIZENESW,
GCursor::ResizeAll = SDL_SYSTEM_CURSOR_SIZEALL,
GCursor::NoAction = SDL_SYSTEM_CURSOR_NO,
GCursor::Hand = SDL_SYSTEM_CURSOR_HAND;

/*
bool GMouseEvent::cmdMode() const {
#ifdef __APPLE__
  return (state() & MetaMod) != 0;
#else
  return (state() & CtrlMod) != 0;
#endif
}

bool GKeyEvent::cmdMode() const {
#ifdef __APPLE__
  return (state() & MetaMode) != 0;
#else
  return (state() & CtrlMode) != 0;
#endif
}*/

GString GEvent::keyToName(GKeyCode key){
  return SDL_GetKeyName(key);
}

GKeyCode GEvent::nameToKey(GString const& keyname){
  return SDL_GetKeyFromName(keyname.data());
}

GEventState GEvent::modifierToState(GKeyCode modifier_keycode) {
  switch (modifier_keycode) {   // e.key.keysym.sym SDL virtual key code
    case SDLK_LSHIFT:
    case SDLK_RSHIFT:
      return GEventState::Shift;
    case SDLK_LCTRL:
    case SDLK_RCTRL:
      return GEventState::Ctrl;
    case SDLK_LALT:
    case SDLK_RALT:
      return GEventState::Alt;
    case SDLK_LGUI:
    case SDLK_RGUI:
      return GEventState::Meta;
    case SDLK_MODE:
      return GEventState::AltGr;
    case SDLK_CAPSLOCK:
      return GEventState::Caps;
    case SDLK_NUMLOCKCLEAR:
      return GEventState::Num;
    default:
      return GEventState::NoState;
  }
}

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

GPoint GNatContext::cursorPos() const {
  int x{}, y{};
  SDL_GetGlobalMouseState(&x, &y); // cf. SDL_GetMouseState(&x, &y);
  return GPoint(float(x), float(y));
}

#ifndef __APPLE__

void GNatContext::moveCursor(GPoint const& pos) {
  SDL_WarpMouseGlobal((int)pos.x, (int)pos.y);  // buggy on Cocoa
}

void GNatContext::showCursor(bool state) {
  SDL_ShowCursor(state ? SDL_ENABLE : SDL_DISABLE);
}

void GNatContext::freezeCursor(bool state) {
  gerror("GNatApp::freezeCursor","Not implemented for this OS");
}

#endif

void GNatContext::setCursorShape(GCursor::Shape type) {
  static SDL_Cursor* curs{};
  static GCursor::Shape curstype{};

  if (type == curstype) return;
  curstype = type;
  SDL_FreeCursor(curs);
  curs = SDL_CreateSystemCursor((SDL_SystemCursor)type);
  SDL_SetCursor(curs);
}

void GNatContext::grabCursor(bool state) {
  SDL_CaptureMouse(state ? SDL_TRUE : SDL_FALSE);
  //return SDL_CaptureMouse(state ? SDL_TRUE : SDL_FALSE) == 0;
}

uint32_t GNatContext::elapsedTime() {
  return SDL_GetTicks();
}

void GNatContext::setSwapInterval(int interval) {
  SDL_GL_SetSwapInterval(interval);
}

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

static inline void bindWindow(SDL_Window* sdlw, GWindow* win) {
  SDL_SetWindowData(sdlw, "gwin", win);
}

static inline GWindow* idToWindow(Uint32 winID) {
  if (SDL_Window* sdlw = SDL_GetWindowFromID(winID))
    return (GWindow*)SDL_GetWindowData(sdlw, "gwin");
  else return nullptr;
}


GNatContext::GNatContext(GControlImpl& ctrl) :
ctrl_(ctrl)
{
  atexit(SDL_Quit);

  //if (SDL_Init(SDL_INIT_EVERYTHING) != 0) {
  if (SDL_Init(SDL_INIT_VIDEO|SDL_INIT_EVENTS|SDL_INIT_TIMER) != 0) {  // SDL_INIT_JOYSTICK
    gerror("GNatApp", "Fatal Error: Can't initialize SDL");
    //SDL_Quit();
    exit(1);
  }
  /*
   SDL_version compiled;
   SDL_version linked;
   SDL_VERSION(&compiled);
   SDL_GetVersion(&linked);
   printf("We compiled against SDL version %d.%d.%d ...\n",
   compiled.major, compiled.minor, compiled.patch);
   printf("But we are linking against SDL version %d.%d.%d.\n",
   linked.major, linked.minor, linked.patch);
   */

  if (TTF_Init() != 0) {
    gerror("GNatApp", "Warning: Can't initialize SDL image module");
  }

  //if ((IMG_Init(IMG_INIT_PNG) & IMG_INIT_PNG) != IMG_INIT_PNG) {
  //  error("natapp",  "IMG_Init");
  //}

  // Set the scaling quality to nearest-pixel
  //if (SDL_SetHint(SDL_HINT_RENDER_SCALE_QUALITY, "0") < 0) {
  if (SDL_SetHint(SDL_HINT_RENDER_SCALE_QUALITY, "1") < 0) {
    gerror("GNatApp", "Warning: Failed to set Render Scale Quality");
  }

  SDL_GL_SetAttribute(SDL_GL_ACCELERATED_VISUAL, 1);
  SDL_GL_SetAttribute(SDL_GL_DOUBLEBUFFER, 1);
  SDL_GL_SetAttribute(SDL_GL_STEREO, 0);
  SDL_GL_SetAttribute(SDL_GL_RED_SIZE, 8);
  SDL_GL_SetAttribute(SDL_GL_GREEN_SIZE, 8);
  SDL_GL_SetAttribute(SDL_GL_BLUE_SIZE, 8);
  SDL_GL_SetAttribute(SDL_GL_ALPHA_SIZE, 8);
  SDL_GL_SetAttribute(SDL_GL_DEPTH_SIZE, 24);
  SDL_GL_SetAttribute(SDL_GL_STENCIL_SIZE, 8);  // !!!
  SDL_GL_SetAttribute(SDL_GL_ACCUM_RED_SIZE, 0);
  SDL_GL_SetAttribute(SDL_GL_ACCUM_GREEN_SIZE, 0);
  SDL_GL_SetAttribute(SDL_GL_ACCUM_BLUE_SIZE, 0);
  SDL_GL_SetAttribute(SDL_GL_ACCUM_ALPHA_SIZE, 0);

  //SDL_GL_SetAttribute(SDL_GL_MULTISAMPLEBUFFERS, 1);
  //SDL_GL_SetAttribute(SDL_GL_MULTISAMPLESAMPLES, 4);

  // share GL contexts between windows
  SDL_GL_SetAttribute(SDL_GL_SHARE_WITH_CURRENT_CONTEXT, 1);

  //SDL_HINT_MAC_CTRL_CLICK_EMULATE_RIGHT_CLICK
  //SDL_EventState(SDL_SYSWMEVENT, SDL_ENABLE);
  //SDL_HINT_MOUSE_NORMAL_SPEED_SCALE
  //SDL_HINT_MOUSE_RELATIVE_SPEED_SCALE

  // To enable drag&drop on your SDL app, you must also edit your info.plist file.
  // Add/Modify Document Types. For example, to enable all document types,
  // add the "public.data" mime type as a document type.
  SDL_EventState(SDL_DROPFILE, SDL_ENABLE);
  // dont work?
  SDL_EventState(SDL_DROPTEXT, SDL_ENABLE);
  SDL_EventState(SDL_DROPBEGIN, SDL_ENABLE);
  SDL_EventState(SDL_DROPCOMPLETE, SDL_ENABLE);

  //SDL_SetHint(SDL_HINT_TOUCH_MOUSE_EVENTS, "0");
  /* Touch Events:
   cerr <<  SDL_GetNumTouchDevices() << endl;
   const char* s = SDL_GetError();
   cerr << s << endl;
   cerr <<  SDL_GetTouchDevice(0) << endl;
   s = SDL_GetError();
   cerr << s << endl;
   cerr <<  SDL_GetTouchDevice(1) << endl;
   //SDL_GetTouchDeviceType()

   SDL_EventState(SDL_FINGERMOTION, SDL_ENABLE);
   SDL_EventState(SDL_FINGERUP, SDL_ENABLE);
   SDL_EventState(SDL_FINGERDOWN, SDL_ENABLE);
   SDL_EventState(SDL_MULTIGESTURE, SDL_ENABLE);


   The most common reason to access a touch device is to normalize inputs.
   This would be accomplished by code like the following:
   SDL_Touch* inTouch = SDL_GetTouch(event.tfinger.touchId);
   if(inTouch == NULL) continue; //The touch has been removed
   float x = ((float)event.tfinger.x)/inTouch->xres;
   float y = ((float)event.tfinger.y)/inTouch->yres;
   */

  /*
   SDL_GetDisplayDPI();
   Create your window with the SDL_WINDOW_ALLOW_HIGHDPI flag, and then use
   SDL_GL_GetDrawableSize() to find the actual drawable size.
   You are responsible for scaling mouse and drawing coordinates appropriately.
   */
  SDL_Rect r{};
  SDL_GetDisplayBounds(0, &r);
  screensize_ = {float(r.w), float(r.h)};
  //cout << "Display: " << r.w << " x " << r.h << endl;
  /*
   {
   int display_in_use = 0;    // using first display
   int i, display_mode_count;
   SDL_DisplayMode mode;
   Uint32 f;
   cout << "SDL_GetNumVideoDisplays " << SDL_GetNumVideoDisplays() << endl;
   display_mode_count = SDL_GetNumDisplayModes(display_in_use);
   if (display_mode_count < 1) {
   cout << "NumDisplayModes failed: " << SDL_GetError() << endl;
   }
   else {
   cout << "NumDisplayModes: " << display_mode_count << endl;
   for (i = 0; i < display_mode_count; ++i) {
   if (SDL_GetDisplayMode(display_in_use, i, &mode) != 0) {
   cout << "SDL_GetDisplayMode failed: " << SDL_GetError() << endl;
   }
   else {
   f = mode.format;
   cout
   << "Mode " << i
   << ":  " << SDL_BITSPERPIXEL(f) << "bpp"
   << " " << SDL_GetPixelFormatName(f)
   << "  " << mode.w << " x " << mode.h << " pixels"
   <<endl;
   }}}}
   */
  SDL_EnableScreenSaver();
}

GNatContext::~GNatContext() {
  TTF_Quit();
  IMG_Quit();
  SDL_Quit();
}


void GNatContext::setMainLoopDelay(unsigned int val) {
  mainloop_delay_ = val;
}

int GNatContext::waitEvent(SDL_Event * event) {
  while (true) {
    SDL_PumpEvents();
    switch (SDL_PeepEvents(event, 1, SDL_GETEVENT, SDL_FIRSTEVENT, SDL_LASTEVENT)) {
      case -1:
        return 0;
      case 0:
        SDL_Delay(mainloop_delay_);  // delay before taking care of the next event
        break;
      default:
        /* Has events */
        return 1;
    }
  }
}

void GNatContext::mainLoop() {
  SDL_Event e;
  while (true) {    // NOTE: update() ne fait pas paint sans event!
    waitEvent(&e);
    processEvent(e);
    ctrl_.paintAll();
  }
}

void GNatContext::subLoop(const bool& subquit) {
  SDL_Event e;
  while (true) {    // NOTE: update() ne fait pas paint sans event!
    waitEvent(&e);
    processEvent(e);
    if (subquit) break;
    ctrl_.paintAll();
  }
}

void GNatContext::processEvent(const SDL_Event& e) {
  static GWindow* mousedown_win{};
  /*
   cerr << e.type
   << " / MOUSE motion " << SDL_MOUSEMOTION
   << " down " << SDL_MOUSEBUTTONDOWN
   << " up " << SDL_MOUSEBUTTONUP
   << " / FINGER motion " << SDL_FINGERMOTION
   << " down " << SDL_FINGERDOWN
   << " up " << SDL_FINGERUP
   << endl;
   */
  
  switch (e.type) {
      
    case SDL_USEREVENT: {
      if (e.user.code == GControlImpl::TimerEvent) {
        auto* timer = (GTimer*)e.user.data1;
        timer->timerCB();
      }
      else if (e.user.code == GControlImpl::FunEvent) {
        if (auto* fun = static_cast<GFunction*>(e.user.data1)) {
          (*fun)();
          ctrl_.funCallDone();
        }
        //else cerr << "# just update" << endl;
      }
      return;
    }

    case SDL_MOUSEMOTION:
      if (auto* win = idToWindow(e.motion.windowID)) {
        // cerr << "MOUSE MOTION " << e.button.windowID << " " << natwin
        // <<" "<< e.button.which << " / " << e.motion.x << " "<< e.motion.y
        // << endl;
        GMouseEvent ge(GTrigger::mousemove, e.motion.timestamp);
        ge.window_ = win;
        ge.button_ = 0;
        if (e.button.which == SDL_TOUCH_MOUSEID) ge.channel_ = 0;  // !!!
        else ge.channel_ = e.motion.which;
        ge.wpos_ = {float(e.motion.x), float(e.motion.y)};
        ge.dx_ = float(e.motion.xrel);
        ge.dy_ = float(e.motion.yrel);
        ctrl_.mouseMove(ge);
      }
      return;

    case SDL_MOUSEBUTTONDOWN:
      /*
      cerr << "\n##MOUSEDOWN win: " << e.button.windowID
      << " / t: " << e.button.timestamp <<endl;
       <<" which: "<< e.button.which
       << " / x: " << e.button.x << " y: " << e.button.y << " "
       << " / b: " << int(e.button.button) << " c: " << int(e.button.clicks)
       << " / t: " << e.button.timestamp <<endl;
       */
      if (auto* win = idToWindow(e.button.windowID)) {
        mousedown_win = win;  // used by SDL_MOUSEBUTTONUP
        GMouseEvent ge(GTrigger::mousedown, e.button.timestamp);
        ge.window_ = win;
        ge.button_ = e.button.button;  // ie 1, 2, 3, etc.
        if (e.button.which == SDL_TOUCH_MOUSEID) ge.channel_ = 0;  // !!!
        else ge.channel_ = e.button.which;
        ge.wpos_ = {(float)e.button.x, (float)e.button.y};
        ctrl_.mouseDown(ge);
      }
      return;

    case SDL_MOUSEBUTTONUP: {
      auto* win = idToWindow(e.button.windowID);
      /*
      cerr << "##MOUSEUP win: " << e.button.windowID
      << " / t: " << e.button.timestamp << endl;
      <<" which: "<< e.button.which
       << " / x: " << e.button.x << " y: " << e.button.y << " "
       << " / b: " << int(e.button.button) << " c: " << int(e.button.clicks)
       */
      if (win) {
        auto time = e.button.timestamp;

        // !! SDL BUG: time is incorrect after a focus gained, with the consequence
        // that wrong double-clicks are detected!
        if (e.button.windowID == focus_winid_
            && e.button.timestamp - focus_time_ < 500) {
          //cerr << "focustime: "<<focus_time_<< " "<< (e.button.timestamp-focus_time_) << endl;
          time -= 200;
          focus_time_ = 0;
        }

        GMouseEvent ge(GTrigger::mouseup, time);  // time: see above
        ge.window_ = win;
        ge.button_ = e.button.button;  // ie 1, 2, 3, etc.
        if (e.button.which == SDL_TOUCH_MOUSEID) ge.channel_ = 0;
        else ge.channel_ = e.button.which;
        ge.wpos_ = {(float)e.button.x, (float)e.button.y};
        ctrl_.mouseUp(ge);
      }

      else /* (!win) */ {   // !! SDL BUG: windowID lost after a mouse leave!
        int x = 0, y = 0;
        Uint32 btn_state = SDL_GetMouseState(&x,&y);
        if (mousedown_win && btn_state == 0) {
          GMouseEvent ge(GTrigger::mouseup, e.button.timestamp);
          ge.window_ = mousedown_win;
          ge.button_ = 0;  // !!!
          if (e.button.which == SDL_TOUCH_MOUSEID) ge.channel_ = 0;
          else ge.channel_ = e.button.which;
          ge.wpos_ = {(float)x, (float)y};
          ctrl_.mouseUp(ge);
        }
      }
      mousedown_win = nullptr;
    } return;

    case SDL_MOUSEWHEEL:
      if (auto* win = idToWindow(e.wheel.windowID)) {
        GMouseEvent ge(GTrigger::wheel, e.wheel.timestamp);
        ge.window_ = win;
        if (e.wheel.direction == SDL_MOUSEWHEEL_FLIPPED) {
          ge.dx_ = float(-e.wheel.x);
          // inversion bizarre: bug SDL avec Cocoa ?
          ge.dy_ = float(+e.wheel.y);
        }
        else {
          ge.dx_ = float(+e.wheel.x);
          ge.dy_ = float(-e.wheel.y);
        }
        ctrl_.wheelMove(ge);
      }
      return;

    case SDL_FINGERMOTION:
    case SDL_FINGERDOWN:
    case SDL_FINGERUP: {
      cerr << "SDL_FINGER "
      << e.tfinger.touchId << " " << e.tfinger.fingerId << " / "
      << e.tfinger.x << " "<< e.tfinger.y << " / "
      << e.tfinger.dx << " "<< e.tfinger.dy
      << endl;
      GTouchEvent ge(GTrigger::touch, e.tfinger.timestamp);
      ge.window_ = ctrl_.mainWindow();    // sent to the main window!
      //ge.channel_ = 0;  // !!!
      if (e.type == SDL_FINGERMOTION) ge.type_ = GTouchEvent::Motion;
      else if (e.type == SDL_FINGERDOWN) ge.type_ = GTouchEvent::Down;
      else if (e.type == SDL_FINGERUP) ge.type_ = GTouchEvent::Up;
      else return;
      ge.touchID_  = e.tfinger.touchId;
      ge.fingerID_ = e.tfinger.fingerId;
      ge.pos_  = {e.tfinger.x, e.tfinger.y};
      ge.dist_ = {e.tfinger.dx, e.tfinger.dy};
      ge.pressure_ = e.tfinger.pressure;
      ctrl_.touchEvent(ge);
    } return;

    case SDL_MULTIGESTURE: {
      /*
       cerr << "SDL_MULTIGESTURE "
       << e.mgesture.touchId << " " << e.mgesture.numFingers << " / "
       << e.mgesture.x << " "<< e.mgesture.y << " / "
       << e.mgesture.dDist << " "<< e.mgesture.dTheta
       << endl;
       */
      GGestureEvent ge(GTrigger::gesture, e.mgesture.timestamp);
      ge.window_ = ctrl_.mainWindow();
      ge.touchID_ = e.mgesture.touchId;
      ge.numfingers_ = e.mgesture.numFingers;
      ge.pos_ = {e.mgesture.x, e.mgesture.y};
      ge.dist_ = e.mgesture.dDist;
      ge.theta_ = e.mgesture.dTheta;
      ctrl_.gestureEvent(ge);
    } return;

    case SDL_KEYDOWN:
      //cerr<<"Keydown " << e.key.keysym.sym <<" " <<endl;
      if (auto* win = idToWindow(e.key.windowID)) {
        GKeyEvent ge(GTrigger::keydown, e.key.timestamp);
        ge.window_ = win;
        ge.key_ = (GKeyCode)e.key.keysym.sym;  // SDL virtual key code
        ctrl_.keyDown(ge);
      }
      return;

    case SDL_KEYUP:
      //cerr<<"Keyup " << e.key.keysym.sym <<endl<<endl;
      if (auto* win = idToWindow(e.key.windowID)) {
        GKeyEvent ge(GTrigger::keyup, e.key.timestamp);
        ge.window_ = win;
        ge.key_ = (GKeyCode)e.key.keysym.sym;  // SDL virtual key code
        ctrl_.keyUp(ge);
      }
      return;

    case SDL_TEXTINPUT:
      //cerr<<"Keytype " << e.text.text <<endl;
      if (auto* win = idToWindow(e.text.windowID)) {
        GKeyEvent ge(GTrigger::keytype, e.key.timestamp);
        ge.window_ = win;
        ge.text_ = e.text.text;
        ctrl_.keyType(ge);
      }
      return;

    case SDL_TEXTEDITING:
      return;

    case SDL_KEYMAPCHANGED:
      //cerr << "SDL_KEYMAPCHANGED" << endl;
      return;

    case SDL_WINDOWEVENT:
      if (auto* win = idToWindow(e.window.windowID)) processWindowEvent(e, *win);
      return;

    case SDL_SYSWMEVENT:
      //i f (e.syswm.msg) {
      // cerr << "SDL_SYSWMEVENT " << e.syswm.msg->subsystem
      // << " " << SDL_SYSWM_COCOA << endl;
      // }
      return;

    case SDL_DROPFILE:
      // see DL_EventState(SDL_DROPFILE, SDL_ENABLE) above
      if (char* filedir = e.drop.file) {
        cerr << "DROPFILE " << filedir << endl;
        SDL_free(filedir);
      }
      return;

    case SDL_DROPTEXT:
      if (char* filedir = e.drop.file) {
        cerr << "DROPTEXT " << filedir << endl;
        SDL_free(filedir);
      }
      return;

    case SDL_DROPBEGIN:
      if (char* filedir = e.drop.file) {
        cerr << "DROPBEGIN" << endl;
        SDL_free(filedir);
      }
      return;

    case SDL_DROPCOMPLETE:
      if (char* filedir = e.drop.file) {
        cerr << "DROPCOMPLETE"<< endl;
        SDL_free(filedir);
      }
      return;

      /*
       case SDL_DOLLARGESTURE:
       cerr << "SDL_DOLLARGESTURE " << e.tfinger.fingerId << endl;
       return;

       case SDL_DOLLARRECORD:
       cerr << "SDL_DOLLARRECORD " << e.tfinger.fingerId << endl;
       return;

       case SDL_JOYAXISMOTION:
       cerr << "SDL_JOYAXISMOTION " << e.tfinger.fingerId << endl;
       return;

       case SDL_JOYBALLMOTION:
       cerr << "SDL_JOYBALLMOTION " << e.tfinger.fingerId << endl;
       return;

       case SDL_JOYHATMOTION:
       cerr << "SDL_JOYHATMOTION " << e.tfinger.fingerId << endl;
       return;

       case SDL_JOYBUTTONDOWN:
       cerr << "SDL_JOYBUTTONDOWN " << e.tfinger.fingerId << endl;
       return;

       case SDL_JOYBUTTONUP:
       cerr << "SDL_JOYBUTTONUP " << e.tfinger.fingerId << endl;
       return;

       case SDL_JOYDEVICEADDED:
       cerr << "SDL_JOYDEVICEADDED " << e.tfinger.fingerId << endl;
       return;

       case SDL_JOYDEVICEREMOVED:
       cerr << "SDL_JOYDEVICEREMOVED " << e.tfinger.fingerId << endl;
       return;

       case SDL_AUDIODEVICEADDED:
       cerr << "SDL_AUDIODEVICEADDED" << endl;
       return;

       case SDL_AUDIODEVICEREMOVED:
       cerr << "SDL_AUDIODEVICEREMOVED" << endl;
       return;

       case SDL_RENDER_TARGETS_RESET:
       cerr << "SDL_RENDER_TARGETS_RESET" << endl;
       return;

       case SDL_RENDER_DEVICE_RESET:
       cerr << "SDL_RENDER_DEVICE_RESET" << endl;
       return;
       */

    case SDL_QUIT:
      GApp::close();
      return;

    default:
      //cerr << "Other event: type: " << e.type << endl;
      return;
  }
}

void GNatContext::pushUserEvent(GFunction* fun) {
  SDL_Event event;
  SDL_UserEvent userevent;
  //cerr << "GNatContext::pushUserEvent"<<endl;
  userevent.type = SDL_USEREVENT;
  userevent.code = GControlImpl::FunEvent;
  userevent.data1 = fun;
  userevent.data2 = nullptr;
  event.type = SDL_USEREVENT;
  event.user = userevent;
  SDL_PushEvent(&event);   // sends the event to mainloop
}

static Uint32 timerCallback(Uint32 delay, void* t) {
  //cerr << "timerCallback "<< GNatContext::elapsedTime() <<endl;
  GTimer* timer = static_cast<GTimer*>(t);
  // if the timer thread is going faster than the event thread, timerCB() has
  // not yet been called to execute previous action(s) and waitingcalls_ then
  // counts how many times timerCB() should have been called. Otherwise, if
  // all events can be processed normally, waitingcalls_ value is either 0
  // (just after calling timerCB) or 1 (when timerCB needs to be called).
  ++timer->waitingcalls_;

  if (timer->waitingcalls_ == 1) {
    SDL_Event event;
    SDL_UserEvent userevent;
    userevent.type = SDL_USEREVENT;
    userevent.code = GControlImpl::TimerEvent;
    userevent.data1 = timer;
    userevent.data2 = nullptr;
    event.type = SDL_USEREVENT;
    event.user = userevent;
    SDL_PushEvent(&event);    // sends the event to mainloop
  }
  // timerCB() have not yet been called because the event thread can not
  // process actions as fast as expected. The next timeout is then cancelled
  // to avoid making things worse and also because requesting a new timeout
  // before executing timerCB() would be unsafe as timerCB() might delete the
  // timer (the function would then be called in the future with a pointer
  // to a timer that don't still exist).
  auto now = GNatContext::elapsedTime();
  auto derive = now - timer->nexttime_;
  auto ajusteddelay = timer->delay() - derive;
  if (ajusteddelay <= 0) ajusteddelay = 1;  // 0 would cancel the timer!
  //cerr << "ajusteddelay: " << ajusteddelay <<endl;
  timer->nexttime_ = now + ajusteddelay;
  //if (acdelay.count() == 0) GApp::error("SDL", "Warning: acdelay.count() is null");
  //return Uint32(acdelay.count());
  return ajusteddelay;
}

bool GNatContext::startTimer(GTimer& timer, GTime delay) {
  //::cerr << "startTimer "<< GNatContext::elapsedTime() <<endl;
  stopTimer(timer);
  SDL_TimerID id = SDL_AddTimer(Uint32(delay), timerCallback, &timer);
  if (id == 0) {
    gerror("SDL", "Could not add timer");
    return false;
  }
  /*
   auto now = chrono::steady_clock::now();
   chrono::milliseconds dur(Uint32(timer.delay()));
   */
  timer.nexttime_ = elapsedTime() + timer.delay();
  timer.nattimer_ = (GNatTimer*)new SDL_TimerID(id);
  return true;
}

bool GNatContext::stopTimer(GTimer& timer) {
  //cerr << "stopTimer "<< &timer
  // <<" "<< (timer->nattimer_ ? *(SDL_TimerID*)timer->nattimer_ : 0) << endl;
  timer.waitingcalls_ = 0;
  if (!timer.nattimer_) return false;
  bool stat = SDL_RemoveTimer(*(SDL_TimerID*)timer.nattimer_);
  if (!stat) GApp::error("SDL", "Could not remove timer");

  delete (SDL_TimerID*)timer.nattimer_;
  timer.nattimer_ = nullptr;
  return stat;
}

bool GNatContext::getClipboard(GString& text) const {
  char * s = SDL_GetClipboardText();
  if (!s) return false;
  else {
    text = s;
    SDL_free(s);
    return true;
  }
}

bool GNatContext::setClipboard(GString const& text) {
  return SDL_SetClipboardText(text.c_str()) == 0;
}

void GNatContext::processWindowEvent(SDL_Event const& e, GWindow& win) {
  switch (e.window.event) {
    case SDL_WINDOWEVENT_SHOWN:
      //cerr << "SDL_WINDOWEVENT_SHOWN " << e.window.windowID << endl;
      ctrl_.windowShown(win, e.window.timestamp);
      win.natwin_->paint();
      return;
    case SDL_WINDOWEVENT_HIDDEN:
      ctrl_.windowHidden(win, e.window.timestamp);
      return;
    case SDL_WINDOWEVENT_EXPOSED:
      //cerr << "EXPOSED :" << e.window.windowID << endl;
      win.natwin_->paint();
      return;
    case SDL_WINDOWEVENT_MOVED:
      //cerr << "MOVED: " << e.window.data1 <<" "<< e.window.data2 << endl;
      win.natwin_->usermove_ = true;
      ctrl_.windowMoved(win, e.window.timestamp, e.window.data1, e.window.data2);
      return;
    case SDL_WINDOWEVENT_SIZE_CHANGED:
      // window size has changed, either as a result of an API call or through
      // the system or user changing the window size
      // Followed by SDL_WINDOWEVENT_RESIZED if the size was changed by an external
      // event, i.e. the user or the window manager
      //cerr << "SIZE_CHANGED: " << e.window.data1 <<" "<< e.window.data2 << endl;
      win.natwin_->resizeCB(e.window.timestamp, e.window.data1, e.window.data2);
      return;
    case SDL_WINDOWEVENT_RESIZED:
      // window has been resized; preceded by SDL_WINDOWEVENT_SIZE_CHANGED
      //cerr << "RESIZED :" << e.window.data1 <<" "<< e.window.data2 << endl;
      win.natwin_->resizeCB(e.window.timestamp, e.window.data1, e.window.data2);
      return;
    case SDL_WINDOWEVENT_MINIMIZED:
      ctrl_.windowMinimized(win, e.window.timestamp);
      return;
    case SDL_WINDOWEVENT_MAXIMIZED:
      ctrl_.windowMaximized(win, e.window.timestamp);
      return;
    case SDL_WINDOWEVENT_RESTORED:
      //cerr<<"RESTORED: "<< e.window.windowID<<endl;
      return;
    case SDL_WINDOWEVENT_ENTER:
      //cerr << "SDL_WINDOWEVENT_ENTER "<< e.window.windowID << endl;
      ctrl_.windowEntered(win, e.window.timestamp);
      return;
    case SDL_WINDOWEVENT_LEAVE:
      //cerr << "SDL_WINDOWEVENT_LEAVE "<< e.window.windowID << endl;
      ctrl_.windowExited(win, e.window.timestamp);
      return;
    case SDL_WINDOWEVENT_FOCUS_GAINED:
      // cerr << "\n##FOCUS_GAINED win: "<< e.window.windowID
      // << " " << e.window.timestamp <<endl;
      focus_winid_ = e.window.windowID;
      focus_time_ = e.window.timestamp;
      ctrl_.windowfocusIn(win, e.window.timestamp);
      return;
    case SDL_WINDOWEVENT_FOCUS_LOST:
      // cerr << "\n##FOCUS_LOST "<< e.window.windowID
      // << " " << GNatContext::elapsedTime()<<endl;
      ctrl_.windowfocusOut(win, e.window.timestamp);
      return;
    case SDL_WINDOWEVENT_CLOSE:
      //cerr << "Window %d closed", e.window.windowID);
      ctrl_.windowClosed(win, e.window.timestamp);
      return;
#if SDL_VERSION_ATLEAST(2, 0, 5)
    case SDL_WINDOWEVENT_TAKE_FOCUS:
      //cerr << "Window is offered a focus "<< e.window.windowID<< endl;
      return;
    case SDL_WINDOWEVENT_HIT_TEST:
      //cerr << "Window has a special hit test " << e.window.windowID<<endl;
      return;
#endif
    default:
      //cerr << "Unknown Window event " << e.window.windowID <<" "<< e.window.event<<endl;
      return;
  }
}

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

GNatWindow::GNatWindow(GWindow* w, bool with_border, bool resizable) :
window_(w) {
  // On Apple's OS X you must set the NSHighResolutionCapable Info.plist
  // property to YES, otherwise you will not receive a High DPI OpenGL canvas.
  Uint32 flags = SDL_WINDOW_OPENGL| SDL_WINDOW_HIDDEN;
  // | SDL_WINDOW_ALLOW_HIGHDPI

  if (resizable) flags |= SDL_WINDOW_RESIZABLE;
  if (!with_border) flags |= SDL_WINDOW_BORDERLESS;
  sdlwin_ = SDL_CreateWindow("",
                             SDL_WINDOWPOS_UNDEFINED, SDL_WINDOWPOS_UNDEFINED,
                             //SDL_WINDOWPOS_CENTERED, SDL_WINDOWPOS_CENTERED,
                             //0, 0,
                             10, 10,
                             flags);
  if (!sdlwin_) {
    gerror("SDL", "Can't create window");
    return;
  }
  glcontext_ = SDL_GL_CreateContext(sdlwin_);
  if (!glcontext_) {
    gerror("SDL", "Can't create GL context");
    return;
  }
  bindWindow(sdlwin_, window_);
  winid_ = SDL_GetWindowID(sdlwin_);
  glClearColor(0.0f, 0.0f, 0.0f, 0.0f);
  // creates the Cocoa delegate (solves a bug when resizing windows)
#ifdef __APPLE__
  createDelegate();
#endif
}

GNatWindow::~GNatWindow() {
  SDL_DestroyWindow(sdlwin_);
#ifdef __APPLE__
  deleteDelegate();
#endif
}

void GNatWindow::move(float x, float y) {
  // this test avoids SDL_WINDOWEVENT_MOVED to call GNatWindow::move(()
  if (!usermove_) {
    usermove_ = false;
    //cerr << "GNatWindow::move " <<window_<<": "<< x <<" "<< y << endl;
    SDL_SetWindowPosition(sdlwin_, (int)x, (int)y);
  }
}

void GNatWindow::resize(float w, float h) {
  // this test avoids SDL_WINDOWEVENT_SIZE_CHANGED and windowDidResize
  // to call GNatWindow::resize(()
  if (!userresize_) {
    userresize_ = false;
    //The window size in screen coordinates may differ from the size in pixels,
    //if the window was created with SDL_WINDOW_ALLOW_HIGHDPI on a platform
    //with high-dpi support (e.g. iOS or OS X). Use SDL_GL_GetDrawableSize()
    //or SDL_GetRendererOutputSize() to get the real client area size in pixels.
    //cerr << "GNatWindow::resize " <<window_<<": "<< w <<" "<< h << endl;
    SDL_SetWindowSize(sdlwin_, int(w * window_->scale_), int(h * window_->scale_));
  }
}

void GNatWindow::resizeCB(uint32_t time, int w, int h) {
  // This may differ from SDL_GetWindowSize() if we're rendering to a high-DPI drawable,
  // i.e. the window was created with SDL_WINDOW_ALLOW_HIGHDPI on a platform
  // with high-DPI support (Apple calls this "Retina"), and not disabled by
  // the SDL_HINT_VIDEO_HIGHDPI_DISABLED hint.

  userresize_ = true;

  SDL_GL_GetDrawableSize(sdlwin_, &w, &h);
  //cerr << "GNatWindow::resizeCB "  <<window_<<" : " << w <<" "<< h << endl;
  window_->frame_.w = w / window_->scale_;
  window_->frame_.h = h / window_->scale_;

  SDL_GL_MakeCurrent(sdlwin_, glcontext_);
  /*
   glViewport(0, 0, (GLsizei)w, (GLsizei)h);
   glMatrixMode(GL_PROJECTION);
   glLoadIdentity();
   glOrtho(0., w, h, 0., -1.f, 1.f);  // (0,0) is top left corner
   glMatrixMode(GL_MODELVIEW);
   */
  //glEnable(GL_LINE_SMOOTH);
  //glHint(GL_LINE_SMOOTH_HINT,  GL_NICEST);

  //glEnable(GL_MULTISAMPLE);
  //glHint(GL_MULTISAMPLE_FILTER_HINT_NV, GL_NICEST);

  glEnable(GL_BLEND);
  glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);

  // NOTE:
  // - appel de Cocoa:windowDidResize egalement necessaire sinon la fenetre
  //    se reaffiche pas pendant qu'on la retaille
  // - cette fct est aussi necessaire sinon la taille du layout s'ajuste pas
  // - noter que la taille est sans titre avec cette fonction alors que
  //   windowDidResize a la taille avec avec titre
  //cerr << "GNatWindow::resizeCB " <<  window_->frame_.w <<" "<< window_->frame_.h  << endl;
  GControlImpl::instance().windowResized(*window_, time,
                                     window_->frame_.w, window_->frame_.h);
}

void GNatWindow::setClip(float x, float y, float w, float h) {
  //cerr << "setClip "<< x <<" "<< x<<" / "<< w<<" "<< h << endl;
  if (w < 0.f) w = 0.f;
  if (h < 0.f) h = 0.f;

  // size increazed to avoid not drawing borders (because of antialiasing?)
  //glViewport(x-1, window_->frame_.h-y-h-1, w+2, h+2);
  auto scale = window_->scale_;
  glViewport(GLint((x-1)*scale), GLint((window_->frame_.h -y-h-1)*scale),
             GLsizei((w+2)*scale), GLsizei((h+2)*scale) );
  glMatrixMode(GL_PROJECTION);
  glLoadIdentity();
  glOrtho(x-1, x+w+1, y+h+1, y-1, -1.f, 1.f);
  glMatrixMode(GL_MODELVIEW);
}

void GNatWindow::show(bool state) {
  if (state) SDL_ShowWindow(sdlwin_);
  else SDL_HideWindow(sdlwin_);
}

void GNatWindow::paint() {
  SDL_GL_MakeCurrent(sdlwin_, glcontext_);
  GControlImpl::instance().paintWindow(window_);
  SDL_GL_SwapWindow(sdlwin_);
}

void GNatWindow::makeCurrent() {
  SDL_GL_MakeCurrent(sdlwin_, glcontext_);
}

void GNatWindow::makeCurrent(GLContext* context) {
  SDL_GL_MakeCurrent(sdlwin_, context);
}

void GNatWindow::swapBuffers() {
  SDL_GL_SwapWindow(sdlwin_);
}

bool GNatWindow::setIcon(GString const& filename) {
  if (SDL_Surface* surf = GNatContext::readSurface(filename)) {
    SDL_SetWindowIcon(sdlwin_, surf);
    SDL_FreeSurface(surf);
    return true;
  }
  else return false;
}

void GNatWindow::raise() {
  SDL_RaiseWindow(sdlwin_);
}

void GNatWindow::maximize() {
  SDL_MaximizeWindow(sdlwin_);
}

void GNatWindow::minimize() {
  SDL_MinimizeWindow(sdlwin_);
}

void GNatWindow::restore() {
  SDL_RestoreWindow(sdlwin_);
}

void GNatWindow::setMinSize(float w, float h) {
  SDL_SetWindowMinimumSize(sdlwin_, (int)w, (int)h);
}

void GNatWindow::setMaxSize(float w, float h) {
  SDL_SetWindowMaximumSize(sdlwin_, (int)w, (int)h);
}

void GNatWindow::setTitle(GString const& title) {
  SDL_SetWindowTitle(sdlwin_, title.c_str());
}

/*
void GNatWindow::getTitle(GString& title) {
  title = SDL_GetWindowTitle(sdlwin_);
}*/

void GNatWindow::setResizable(bool resizable){
  SDL_SetWindowResizable(sdlwin_, (resizable ? SDL_TRUE : SDL_FALSE));
}

void GNatWindow::setBordered(bool bordered) {
  SDL_SetWindowBordered(sdlwin_, (bordered ? SDL_TRUE : SDL_FALSE));
}

bool GNatWindow::setOpacity(float opacity) {
  return SDL_SetWindowOpacity(sdlwin_, opacity) == 0;
}

GLContext* GNatWindow::createGLContext() {
  return (GLContext*)SDL_GL_CreateContext(sdlwin_);
}

/*
bool GNatWindow::getSystemInfo(SDL_SysWMinfo& info) {
  SDL_VERSION(&info.version);
  if (SDL_GetWindowWMInfo(sdlwin_, &info)) {
    switch (info.subsystem) {
        
#if defined(SDL_VIDEO_DRIVER_COCOA)
      case SDL_SYSWM_COCOA:
        //cerr <<"Cocoa Window " <<info.info.cocoa.window<<endl;
        break;
#endif
#if defined(SDL_VIDEO_DRIVER_X11)
      case SDL_SYSWM_X11:
        break;
#endif
#if defined(SDL_VIDEO_DRIVER_DIRECTFB)
      case SDL_SYSWM_DIRECTFB:
        break;
#endif
#if defined(SDL_VIDEO_DRIVER_WINDOWS)
      case SDL_SYSWM_WINDOWS:
        //cerr <<"Windows Window " <<info.info.win.window<<endl;
        break;
#endif
#if defined(SDL_VIDEO_DRIVER_WAYLAND)
      case SDL_SYSWM_WAYLAND:
        break;
#endif
      default:
        break;
    }
    return true;
  }
  else return false;
}
*/

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

bool GNatContext::loadFont(GFontResource& f) {
  if (!f.face_) return false;
  f.natfont_ = (GNatFont*) TTF_OpenFont(f.face_->path_.c_str(), f.size_);
  return f.natfont_ != nullptr;
}

void GNatContext::deleteFont(GFontResource& f) {
  if (f.natfont_) TTF_CloseFont((TTF_Font*)f.natfont_);
}

GDim GNatContext::textSize(GFontResource& f, GString const& text) {
  if (!f.natfont_) return GDim();
  if (text.empty()) return GDim(0, fontHeight(f));
  int w{}, h{};
  if (TTF_SizeUTF8((TTF_Font*)f.natfont_, text.c_str(), &w, &h) == 0) return GDim(float(w), float(h));
  else return GDim(0, fontHeight(f));
}

float GNatContext::textWidth(GFontResource& f, GString const& text) {
  if (!f.natfont_ || text.empty()) return 0.f;
  int w{};
  if (TTF_SizeUTF8((TTF_Font*)f.natfont_, text.c_str(), &w, nullptr) == 0) return float(w);
  else return 0.f;
}

float GNatContext::fontHeight(GFontResource& f) {
  if (!f.natfont_) return 0.f;
  else return (float)TTF_FontHeight((TTF_Font*)f.natfont_);
}

float GNatContext::fontAscent(GFontResource& f) {
  if (!f.natfont_) return 0.f;
  return (float)TTF_FontAscent((TTF_Font*)f.natfont_);
}

float GNatContext::fontDescent(GFontResource& f) {
  if (!f.natfont_) return 0.f;
  return (float)TTF_FontDescent((TTF_Font*)f.natfont_);
}

bool GNatContext::getFontInfo(GString const& path, GString& family, GString& style) {
  if (path.empty()) return false;
  auto* f = TTF_OpenFont(path.c_str(), 12);  // arbitrary size for testing
  if (!f) return false;

  family = TTF_FontFaceFamilyName(f);
  style = TTF_FontFaceStyleName(f);
  TTF_CloseFont(f);
  return true;
}

/*
bool GNatContext::fontInfo(GString const& fontpath, int size, GFontInfo& info) {
  if (fontpath.empty()) return false;
  auto* f = TTF_OpenFont(fontpath.c_str(), size);
  if (!f) return false;
  else {
    info.family = TTF_FontFaceFamilyName(f);
    info.style = TTF_FontFaceStyleName(f);
    //info.style1 = (TTF_GetFontStyle(f) & TTF_STYLE_ITALIC) ? 'i' : 0;
    //info.weight = (TTF_GetFontStyle(f) & TTF_STYLE_BOLD) ? 'b' : 0;
    // TTF_FontFaceIsFixedWidth
    // Size
    TTF_CloseFont(f);
    return true;
  }
}*/

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

SDL_Surface* GNatContext::readSurface(GString const& path) {
  if (path.empty()) return nullptr;

  static GString jokers("./\\~:!");

  if (jokers.find(path[0]) != GString::npos) {
    return IMG_Load(normPath(path).c_str());
  }
  else {
    auto p = normPath(path);
    SDL_Surface* surf = IMG_Load(p.c_str());
    if (!surf) {
      for (auto& dir : GApp::resourceDirs()) {
        surf = IMG_Load((dir+p).c_str());
        if (surf) break;
      }
    }
    return surf;
  }
}

SDL_Surface* GNatContext::readSurface(const char** xpm) {
  if (!xpm) return nullptr;
  return IMG_ReadXPMFromArray((char**)xpm);
}

void GNatContext::deleteSurface(SDL_Surface* surf) {
  if (surf) SDL_FreeSurface(surf);
}

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

GLuint GNatContext::surfaceToTexture(SDL_Surface* surf) {  // deletes surf!
  if (!surf) return 0;  // 0 is not a valid texture; glIsTexture() will return false

  if (surf->format->BytesPerPixel == 1) {
    // must be converted to 24bits to be displayed
    SDL_Surface* surf2 = SDL_ConvertSurfaceFormat(surf, SDL_PIXELFORMAT_RGB24, 0);
    SDL_FreeSurface(surf);
    if (!surf2) return 0;
    surf = surf2;
  }

  GLenum format = 0;
  GLint colcount = surf->format->BytesPerPixel;

  if (colcount == 4) {   // contains an alpha channel
    if (surf->format->Rmask==0x000000ff) format = GL_RGBA;
    else format = GL_BGRA;
  }
  else if (colcount == 3) {
    if (surf->format->Rmask==0x000000ff) format = GL_RGB;
    else format = GL_BGR;
  }
  /*
   else if (colcount == 1) {
   cerr
   << surf->format->format<<" "
   << SDL_GetPixelFormatName(surf->format->format)
   <<"\n"
   << SDL_PIXELTYPE(surf->format->format) <<" "
   << SDL_PIXELORDER(surf->format->format) <<" "
   << SDL_PIXELLAYOUT(surf->format->format) <<" "
   << SDL_BITSPERPIXEL(surf->format->format) <<" "
   << SDL_ISPIXELFORMAT_INDEXED(surf->format->format) <<" "
   << SDL_ISPIXELFORMAT_ALPHA(surf->format->format) <<" "
   << surf->format->format <<" "
   << surf->format->palette <<"\n"
   << surf->w <<" " << surf->h <<" "<< surf->pitch <<"\n"
   << int(surf->format->BitsPerPixel) <<" " << int(surf->format->BytesPerPixel) <<"\n"
   << surf->format->Rmask <<" "
   << surf->format->Gmask <<" "
   << surf->format->Bmask <<" "
   << surf->format->Amask <<" "
   <<endl;
   return false;
   }
   */
  else {
    gerror("genTexture","Invalid texture format: color count = "+to_string(colcount));
    SDL_FreeSurface(surf);  // deletes surface!
    return 0;
  }

  GLuint texture{};
  glEnable(GL_TEXTURE_2D);
  glGenTextures(1, &texture);
  glBindTexture(GL_TEXTURE_2D, texture);
  glTexImage2D(GL_TEXTURE_2D, 0, colcount, surf->w, surf->h, 0, format,
               GL_UNSIGNED_BYTE, surf->pixels);
  //glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
  //glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
  //glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
  //glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
  //glGenerateMipmap(GL_TEXTURE_2D);
  glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
  glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
  glDisable(GL_TEXTURE_2D);
  SDL_FreeSurface(surf);  // deletes surface !!!
  return texture;
}

void GNatContext::readTexture(GString const& filename, GIconValue& ima) {
  glEnable(GL_TEXTURE_2D);
  if (ima.texture) {
    glDeleteTextures(1, &ima.texture);
    ima.texture = 0;
  }
  auto* surf = readSurface(filename);
  if (surf) {
    ima.w = surf->w;
    ima.h = surf->h;
    ima.texture = surfaceToTexture(surf); // deletes surf !
    ima.status = GIconValue::Loaded;
  }
  else {
    ima.w = 0;
    ima.h = 0;
    ima.texture = 0;
    ima.status = GIconValue::CantLoad;
  }
  glDisable(GL_TEXTURE_2D);
}

void GNatContext::readTexture(const char** xpm, GIconValue& ima) {
  glEnable(GL_TEXTURE_2D);
  if (ima.texture) {
    glDeleteTextures(1, &ima.texture);
    ima.texture = 0;
  }
  auto* surf = readSurface(xpm);
  if (surf) {
    ima.w = surf->w;
    ima.h = surf->h;
    ima.texture = surfaceToTexture(surf); // deletes surf !
    ima.status = GIconValue::Loaded;
  }
  else {
    ima.w = 0;
    ima.h = 0;
    ima.texture = 0;
    ima.status = GIconValue::CantLoad;
  }
  glDisable(GL_TEXTURE_2D);
}

GTexture GNatContext::createFontTexture(GString const& text, GFontResource& f) {
  if (!f.natfont_) return 0;
  auto* surf = TTF_RenderUTF8_Blended((TTF_Font*)f.natfont_,
                                      text.c_str(), SDL_Color{255, 255, 255, 255});
  return surf ? surfaceToTexture(surf) : 0;    // deletes surf !
}

}
