//
// Alignment and justification of the content of a gadget
//
// Guit GUI Toolkit (https://www.telecom-paris.fr/~elc/guit/)
// Copyright 2020 Eric Lecolinet (eric.lecolinet@gmail.com) - Telecom Paris
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//

#ifndef GuitAlign_hpp
#define GuitAlign_hpp
#include <gvarprop.hpp>
namespace guit {

struct GAlignValue {
  GAlignValue();
  GAlignValue(int8_t main, int8_t cross, int8_t icon);
  bool operator==(GAlignValue const&) const;
  bool operator!=(GAlignValue const&) const;
  int8_t main, cross, icon;
};

inline bool gconvert(GAlignValue const& from, GAlignValue& to) {to = from; return true;}
bool gconvert(GString const& from, GAlignValue& to);
bool gconvert(GAlignValue const& from, GString& to);


///  Alignment and justification of the content of a gadget.
/// The gadget children (text, icon, gadgets if any) is justified according
/// to this property, which has 4 fields:
/// - main field: how the children are justified in its main direction
/// - cross field: same thing in the cross direction
/// - spacing field: spacing (in pixels) between children in the main direction
/// - icon field : position of the icon (if any) in the main direction
///
/// The direction depends on the GFlow prop of the gadget (either present in
/// the proplist or inherited from the gadget style), see GFlow.
///
/// Possible values for main, cross and item fields are:
/// - Start: left side if direction is row, Top side if column
/// - End: right side if direction is row, Bottom side if column
/// - Center: centered in this direction
/// - Expand: stretched in this direction
///
class GAlign : public GVarProp_<GAlignValue> {
public:
  using Super = GVarProp_<GAlignValue>;
  enum Layout : int8_t {Start, Center, End, Extend, Equal};

  GAlign();
  GAlign(Layout main, Layout cross, Layout icon = Start);
  GAlign(GAlign const&);
  GAlign(GAlignValue const&);

  /// constr.
  /// argument is "main cross [icon]" with _main_, _cross_, _icon_  one of:
  /// - "start" or "s" (case insensitive)
  /// - "end" or "e"
  /// - "center" or "c"
  /// - "extend" or "x"
  /// - "equal" or "eq"
  /// See GAlign class for meaning. Ending values can be omitted.
  GAlign(GString const&);

  GAlign* clone(bool copy_value) const override;

  Layout mainLayout() const {return Layout(value_.main);}
  Layout crossLayout() const {return Layout(value_.cross);}
  Layout iconLayout() const {return Layout(value_.icon);}

  GAlign& operator=(GAlign const& val) {set(val); return *this;}
  GAlign& operator=(GString const& val) {set(val); return *this;}

  void setMainLayout(Layout);
  void setCrossLaoyut(Layout);
  void setIconLayout(Layout);

  using GType = GPropType_<GAlign,GProp>;
  static GType& Type();
  GType& type() const override {return Type();}

  void updateGadget(Gadget&) override;
  void apply(Gadget*, GRender&, Specif) final;
};

/// Return a GAlign created in the heap.
/// @{
GAlign& Align();
GAlign& Align(GString const& values);
GAlign& Align(GAlign::Layout main, GAlign::Layout cross,
              GAlign::Layout icon = GAlign::Start);
/// @}

}
#endif
