//
//  order Prop
//
// Guit GUI Toolkit (https://www.telecom-paris.fr/~elc/guit/)
// Copyright 2020 Eric Lecolinet (eric.lecolinet@gmail.com) - Telecom Paris
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//

#include <sstream>
#include <gapp.hpp>
#include <gevent.hpp>
#include <gborder.hpp>
#include <gexpr.hpp>
#include <gbox.hpp>
#include <ggraphics.hpp>
#include <gcontrol.hpp>
#include <grender.hpp>
#include <gstring.hpp>
using namespace std;
namespace guit {

GIMPLEMENT_PROP_TYPE(Border, GPropType::Border)

static GCore::BeginPlainObjects begin;
GBorder GBorder::none(GBorder::None, 0,0, GColor::none, GColor::none);
GBorder GBorder::squared(GBorder::Squared, 4,2, GColor::lightgray, GColor::lightgray);
GBorder GBorder::rounded(GBorder::Rounded, 4,4, GColor::lightgray, GColor::lightgray);
GBorder GBorder::oval(GBorder::Oval, 4,4, GColor::lightgray, GColor::lightgray);
GBorder GBorder::polygon(GBorder::Polygon, 0,0, GColor::lightgray, GColor::lightgray);
GBorder GBorder::textedit(GBorder::TextEdit, 4,4, GColor::lightgray, GColor::guitblue);
static GCore::EndPlainObjects end;


bool GBorderValue::hasDrawShape(Gadget& item) {
  return (style & (Rounded | Oval | Polygon)) != 0;
}

bool GBorderValue::hasPickShape(Gadget& item) {
  return (style & (Oval | Polygon)) != 0;
}

void GBorderValue::drawBorder(GRenderGraphics& g, GRender& r) {
  //g.setBgColor(r.background_);
  g.setThickness(1);

  switch (style) {
    case Squared:
      g.drawRect(g.area_);
      return;

    case Rounded:
      g.drawRoundRect(g.area_, 5.,5.);
      return;

    case GBorderValue::TextEdit:
      if (r.gadget().isFocus()) {
        g.drawRect(g.area_);
        g.drawRect(GRect(g.area_.x+1, g.area_.y+1, g.area_.w-2, g.area_.h-2));
      }
      else g.drawRect(g.area_);
      return;

    case GBorderValue::Oval:
      g.setThickness(1);
      g.drawOval(g.area_);
      return;

    case GBorderValue::Polygon:
      if (polygon) {
        g.translate(g.area_.x, g.area_.y);
        g.drawPolygon(*polygon);
        g.translate(-g.area_.x, -g.area_.y);
      }
      return;

    default:
      return;
  }
}

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

bool GBorderValue::pickBorder(Gadget& item, GRect const& r, GPoint const& pt) {
  if (style == Oval) {
    return GGraphics::pickOval(r, pt);
  }
  if (style == Polygon) {
    return polygon ? polygon->contain(GPoint(pt.x-r.x, pt.y-r.y)) : true;
  }
  auto* par = item.getParent();
  if (par && par->isPie()) {
    return GGraphics::pickPie(r, item.angleStart(), item.angleExtent(), pt);
  }
  return true;
}

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

GBorderValue::~GBorderValue() {
  delete polygon;
}

GBorderValue::GBorderValue(Kind style,
                           float left, float right, float top, float bottom,
                           //DrawFun draw, PickFun pick,
                           GColor& idle, GColor& sel) :
GMarginValue(left, right, top, bottom),
style(style),
idlecolor(idle), activecolor(sel)
{}

GBorderValue::GBorderValue(GPoints const& polygon,
                           GColor& idle, GColor& sel) :
style(Polygon),
idlecolor(idle), activecolor(sel),
polygon(new GPoints(polygon))
{}

GBorderValue::GBorderValue(GBorderValue const& from) :
GMarginValue(from),
style(from.style),
idlecolor(from.idlecolor),
activecolor(from.activecolor),
polygon(from.polygon ? new GPoints(*from.polygon) : nullptr)
{}

GBorderValue& GBorderValue::operator=(GBorderValue const& from) {
  GMarginValue::operator=(from);
  style = from.style;
  idlecolor = from.idlecolor;
  activecolor = from.activecolor;
  polygon = from.polygon ? new GPoints(*from.polygon) : nullptr;
  return *this;
}

bool GBorderValue::operator==(const GBorderValue& other) const {
  return
  style == other.style
  && idlecolor == other.idlecolor && activecolor == other.activecolor
  && left == other.left && right == other.right
  && top == other.top && bottom == other.bottom
  && polygon == other.polygon;
}

bool GBorderValue::operator!=(const GBorderValue& other) const {
  return
  style != other.style
  || idlecolor != other.idlecolor || activecolor != other.activecolor
  || left != other.left || right != other.right
  || top != other.top || bottom != other.bottom
  || polygon != other.polygon;
}

bool gconvert(GString const& from, GBorderValue& to) {
  to.style = GBorderValue::None;
  to.left = to.right = to.top = to.bottom = 0;
  to.idlecolor = GColor::black;  // fallback
  to.activecolor = GColor::black;

  GString curloc = setlocale(LC_ALL,NULL);
  setlocale(LC_NUMERIC, "C");

  float val = 0.f;
  size_t b = 0;
  GStrings v;
  size_t n = strsplit(from, v, "", 0, SplitMode::CheckQuotes); // dont remove quotes
  if (n <= 0) goto END;

  switch (strid(v[0])) {
    case strid("none"):
      to = GBorder::none();
      break;
    case strid("squared"):
      to = GBorder::squared();
      break;
    case strid("rounded"):
      to = GBorder::rounded();
      break;
    case strid("oval"):
      to = GBorder::oval();
      break;
    case strid("textedit"):
      to = GBorder::textedit();
      break;
    case strid("polygon"):
      to = GBorder::polygon();
      break;
    default:
      gerror("Border", "Unknown Border '"+from+"'");
      return false;
  }
  b++;
  if (b >= n) goto END;

  while (b < n) {
    if (v[b][0] < '0' || v[b][0] > '9') {b--; break;}
    b++;
  }
  switch (b) {
    case 1+1:
      to.left = to.right = to.top = to.bottom = gconvert(v[1], val) ? val : 0;
      break;
    case 1+2:
      to.left = to.right = gconvert(v[1], val) ? val : 0;
      to.top  = to.bottom = gconvert(v[2], val) ? val : 0;
      break;
    case 1+4:
      to.left   = gconvert(v[1], val) ? val : 0;
      to.right  = gconvert(v[2], val) ? val : 0;
      to.top    = gconvert(v[3], val) ? val : 0;
      to.bottom = gconvert(v[4], val) ? val : 0;
      break;
    default:
      gerror("Border", "Invalid format");
      return false;
      break;
  }
  b++;

  if (b < n) {
    //to.idlecolor = new GColor(strtrimmed(v[b], true)); // unquote
    if (v[b][0] == '\'' || v[b][0] == '"') strtrim(v[b], true); // unquote
    to.idlecolor = new GColor(v[b]);
    b++;
  }

  if (b < n) {
    //to.selectcolor = new GColor(strtrimmed(v[b], true));
    if (v[b][0] == '\'' || v[b][0] == '"') strtrim(v[b], true); // unquote
    to.activecolor = new GColor(v[b]);
    b++;
  }

END:
  setlocale(LC_NUMERIC, curloc.data());
  return true;
}

bool gconvert(GBorderValue const& from, GString& to) {
  string type;
  switch (from.style) {
    case GBorderValue::None:
      type = "none";
      break;
    case GBorderValue::Squared:
      type = "squared";
      break;
    case GBorderValue::Rounded:
      type = "rounded";
      break;
    case GBorderValue::Oval:
      type = "oval";
      break;
    case GBorderValue::TextEdit:
      type = "textedit";
      break;
    case GBorderValue::Polygon:
      type = "polygon";
      break;
  }
  std::ostringstream ss;
  ss.imbue(std::locale("C"));
  ss.precision(GPrefs::precision());

  if (from.left == from.right && from.top == from.bottom) {
    ss << type <<' '<< from.left <<' '<< from.top;
  }
  else {
    ss << type
    <<' '<< from.left <<' '<< from.right <<' '<< from.top <<' '<< from.bottom;
  }
  if (from.idlecolor) {
    ss << " '" << *from.idlecolor << "'";
    if (from.activecolor) ss << " '" << *from.activecolor << "'";
  }
  to = ss.str();
  return true;
}

std::ostream& operator<<(std::ostream& out, const GBorderValue& val) {
  GString str;
  gconvert(val, str);
  return out << str;
}

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

GBorder::GBorder() = default;

GBorder::GBorder(Kind style, float hpad, float vpad, GColor& idle, GColor& sel)
: GVarProp_<GBorderValue>(GBorderValue(GBorderValue::Kind(style),
                                       hpad,hpad, vpad,vpad, idle,sel)) {
}

GBorder::GBorder(GPoints const& polyg, GColor& idle, GColor& sel)
: GVarProp_<GBorderValue>(GBorderValue(polyg, idle, sel)) {}

GBorder::GBorder(GBorder const& from, float hpad, float vpad, GColor* idle, GColor* sel)
: GVarProp_<GBorderValue>(GBorderValue(from.value_.style, hpad, hpad, vpad, vpad,
                                       (idle ? *idle : *from.value_.idlecolor.get()),
                                       (sel ? *sel : *from.value_.activecolor.get())
                                       )) {
}

GBorder::GBorder(GBorderValue const& value) : Super(value) {}

GBorder::GBorder(GString const& format) {set(format);}

GBorder* GBorder::clone(bool copy_value) const {
  return copy_value ? new GBorder(value_) : new GBorder;
}

void GBorder::updateGadget(Gadget& g) {
  g.updateLayout();
}

void GBorder::apply(Gadget*, GRender& r, Specif specif) {
  if (specif < r.specif[GPropType::Border]) return;
  r.specif[GPropType::Border] = specif;
  r.border_ = &value_;
}

}
