//
// Active Expressions (auto-reevaluated)
//
// Guit GUI Toolkit (https://www.telecom-paris.fr/~elc/guit/)
// Copyright 2020 Eric Lecolinet (eric.lecolinet@gmail.com) - Telecom Paris
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//

#ifndef Guit_Expr_hpp
#define Guit_Expr_hpp
#include <vector>
#include <gnumber.hpp>
#include <gwrapper.hpp>
namespace guit {

/// Boolean argument of a GExpr.
class GExprBool {
public:
  GExprBool(bool);
  GExprBool(GBool&);
  GExprBool(GPropRef_<GBool> const&);
  GExprBool(GBoolExpr&);
  GExprBool(std::function<bool()> const&);
  GProp& prop;
};


/// Textual argument of a GExpr.
class GExprText {
public:
  GExprText(const char*);
  GExprText(GString const&);
  GExprText(GPropRef const&);
  GExprText(GTextExpr&);
  GProp const& prop;
};


/// Numeric argument of a GExpr..
class GExprNumber {
public:
  GExprNumber(int);
  GExprNumber(float);
  GExprNumber(double);
  GExprNumber(GNumPropRef const&);
  GExprNumber(GNumExpr&);
  GExprNumber(GBoolExpr&);
  GProp const& prop;
};


/// Active Expression.
/// Active Expression are reified and automatically reevaluated when the value
/// of one of their mutable props changes. Such expressions can be:
///
/// 1) Added to a gadget. For instance, to convert from Euros to Dollars:
/// @code
///     auto& euros = Float(1);
///     auto& rate = Float(1.12);
///
///     auto& box = VBox
///     << (TextField << euros)
///     << (TextField << rate)
///     << ("Result: " + (~euros * ~rate) + "$");  // reactive expression
///
///     euros = 100;
/// @endcode
/// The last line is a textual reactive expression, which displays
/// "Result: 112$" (i.e. 100 euros * a rate of 1.12)
///
/// It will be automatically updated each time the value of 'euros' or 'rate'
/// is changed. 'euros' or 'rate' must be preceded by a ~.
/// This indicates that the expr depends on these variables
///
/// 2) Serve as a condition in conditional statements. The action will be fired
/// if the RExpr becomes true when the value of its variables changes:
/// @code
///     auto& status = Int(0);  // integer prop
///     auto& label = Label
///     << (~status == 0) / "<color=green>OK"
///     << (~status == 1) / "<color=orange>Warning"
///     << (~status == 2) / "<color=red>DANGER!!!"
///
///     status = 2;
/// @endcode
/// The Label will display "DANGER!!!" in red color. As above, it will be
/// changed if the value of 'status' is changed.
///
/// The active expression can contain variables (mutable props or states) of
/// various types provided that it is syntactically correct and returns a bool.
/// It can contain && and || conjonctors.
/// @see CondProp.
class GExpr : public GProp {
public:
  virtual bool isBoolExpr() const {return false;}
  virtual bool isNumExpr()  const {return false;}
  virtual bool isTextExpr() const {return false;}
  GExpr* toExpr() override {return this;}

  bool    boolValue() const override = 0;
  float   floatValue() const override = 0;
  GString stringValue() const override = 0;

  /// can't be cloned.
  GExpr* clone(bool) const override {return nullptr;}

  static GPropType& Type();
  GPropType& type() const override {return Type();}

  /// Vector element.
  template <class T>
  class VectArg {
    std::vector<T>& vect_;
  public:
    VectArg(std::vector<T>& v) : vect_(v) {}
    auto operator[](GNumPropRef index) {
      return GPropRef(*new GVectorWrapper<T>(index.prop_, vect_));
    }
  };

  /// Numerical vector element.
  template <class T>
  class NumVectArg {
    std::vector<T>& vect_;
  public:
    NumVectArg(std::vector<T>& v) : vect_(v) {}
    auto operator[](GNumPropRef index) {
      return GNumPropRef(*new GNumVectorWrapper<T>(index.prop_, vect_));
    }
  };

  enum Oper {
    Equal, NotEqual, Greater, GreaterEqual, Less, LessEqual,
    And, Or, Not, Identity, Add, Substract, Multiply, Divide,
    StrStart, StrContain, StrRegex
  };

protected:
  friend class GExprWrapper;
  friend class GBoolExpr;
  gptr<GProp> left_, right_;
  Oper oper_;

  GExpr(GProp const* left, Oper oper, GProp const* right);
  GExpr(const GExpr& from);
  
  bool isEquivalent(const GProp& other) const override {return false;} // !!
  void bindImpl(class GExprWrapper&);
  void linkImpl(class GExprTrigger&, class GBoolExpr&, bool trigger_or_guard);
};


/// Vector argument in an Active Expression.
/// (~vect)[~index] can be part of an Expr provided that:
/// - _vect_ is a std::vector and vect[i] a valid argument of an Expr
/// - _index_ a mutable number (typically a GInt prop)
/// @caution _vect_ is NOT copied and should thus NOT be destroyed!
template <class T, std::enable_if_t<std::is_arithmetic<T>::value,int> = 0>
inline GExpr::NumVectArg<T> operator~(std::vector<T>& vect) {
  return GExpr::NumVectArg<T>(vect);
}

template <class T, std::enable_if_t<std::is_compound<T>::value,int> = 0>
inline GExpr::VectArg<T> operator~(std::vector<T>& vect) {
  return GExpr::VectArg<T>(vect);
}

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

// negation
GBoolExpr& operator!(GExprBool const& a);
//GBoolExpr& operator!(BoolExpr const& a);

/// true when both expressions are true.
GBoolExpr& operator&&(GExprBool const& a, GExprBool const& b);
GBoolExpr& operator&&(GExprBool const& a, std::function<bool()> const& fun);

/// true when one of these expressions is true.
GBoolExpr& operator||(GExprBool const& a, GExprBool const& b);
GBoolExpr& operator||(GExprBool const& a, std::function<bool()> const& fun);

GBoolExpr& operator==(GExprBool const& a, GExprBool const& b);
GBoolExpr& operator!=(GExprBool const& a, GExprBool const& b);

GBoolExpr& operator==(GExprNumber const& a, GExprNumber const& b);
GBoolExpr& operator!=(GExprNumber const& a, GExprNumber const& b);
GBoolExpr& operator> (GExprNumber const& a, GExprNumber const& b);
GBoolExpr& operator>=(GExprNumber const& a, GExprNumber const& b);
GBoolExpr& operator< (GExprNumber const& a, GExprNumber const& b);
GBoolExpr& operator<=(GExprNumber const& a, GExprNumber const& b);

GNumExpr& operator+(GExprNumber const& a, GExprNumber const& b);
GNumExpr& operator-(GExprNumber const& a, GExprNumber const& b);
GNumExpr& operator*(GExprNumber const& a, GExprNumber const& b);
GNumExpr& operator/(GExprNumber const& a, GExprNumber const& b);

/// true if both arguments have the same value.
GBoolExpr& operator==(GExprText const& a, GExprText const& b);

/// true if first argument starts with second argument.
GBoolExpr& operator^=(GExprText const& a, GExprText const& begin);

/// true if first argument contains the second argument.
GBoolExpr& operator*=(GExprText const& a, GExprText const& substr);

/// true if first argument matches the regex specified in the second argument.
GBoolExpr& operator%=(GExprText const& s, GExprText const& regex);

GBoolExpr& operator!=(GExprText const& a, GExprText const& b);
GBoolExpr& operator> (GExprText const& a, GExprText const& b);
GBoolExpr& operator>=(GExprText const& a, GExprText const& b);
GBoolExpr& operator< (GExprText const& a, GExprText const& b);
GBoolExpr& operator<=(GExprText const& a, GExprText const& b);

GTextExpr& operator+(GExprText const& a, GExprText const& b);
GTextExpr& operator+(GExprText const& a, GExprNumber const& b);
GTextExpr& operator+(GExprNumber const& a, GExprText const& b);

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

/// Boolean Expression.
/// @See GExpr.
class GBoolExpr : public GExpr {
public:
  enum Type {TBool, TNum, TText};
  GBoolExpr(GProp const* left, Oper oper, GProp const* right, Type);
  GBoolExpr(GBoolExpr const& from);

  bool isBoolExpr() const override {return true;}
  bool isNumber() const override {return true;}
  GBoolExpr* toBoolExpr() override {return this;}

  bool    boolValue()   const override;
  float   floatValue()  const override {return float(boolValue());}
  GString stringValue() const override;

  void linkWithTrigger(class GExprTrigger&);
protected:
  Type type_;
};


/// Textual Expression.
/// @See GExpr.
class GTextExpr : public GExpr {
public:
  GTextExpr(GProp const* left, Oper oper, GProp const* right) : GExpr(left,oper,right) {}
  GTextExpr(GTextExpr const& from) : GExpr(from) {}

  bool isTextExpr() const override {return true;}
  bool isString() const override {return true;}

  bool    boolValue()   const override;
  float   floatValue()  const override;
  GString stringValue() const override;
};


/// Numeric Expression.
/// @See GExpr.
class GNumExpr : public GExpr {
public:
  GNumExpr(GProp const* left, Oper oper, GProp const* right) : GExpr(left,oper,right) {}
  GNumExpr(const GNumExpr& from) : GExpr(from) {}

  bool isNumExpr() const override {return true;}
  bool isNumber() const override {return true;}

  bool    boolValue()   const override {return bool(floatValue());}
  float   floatValue()  const override;
  GString stringValue() const override;
};

}
#endif
