//
// implementation of misc layout properties.
//
// Guit GUI Toolkit (https://www.telecom-paris.fr/~elc/guit/)
// Copyright 2020 Eric Lecolinet (eric.lecolinet@gmail.com) - Telecom Paris
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//

#include <sstream>
#include <gevent.hpp>
#include <gexpr.hpp>
#include <gpropspec.hpp>
#include <gcontrol.hpp>
#include <grender.hpp>
using namespace std;
namespace guit {

GIMPLEMENT_PROP_TYPE(Flow, GPropType::Flow)
GIMPLEMENT_PROP_TYPE(Flex, GPropType::Flex)
GIMPLEMENT_PROP_TYPE(Align, GPropType::Align)
GIMPLEMENT_PROP_TYPE(Gap, GPropType::Gap)

GAlign& Align() {return *new GAlign;}

GAlign& Align(GString const& values) {return *new GAlign(values);}

GAlign& Align(GAlign::Layout main, GAlign::Layout cross, GAlign::Layout icon) {
  return *new GAlign(main, cross, icon);
}

static GCore::BeginPlainObjects begin;
GFlow
GFlow::column(GFlow::Column),
GFlow::row(GFlow::Row),
GFlow::pie(GFlow::Pie),
GFlow::stack(GFlow::Stack);
static GCore::EndPlainObjects end;


bool GFlowValue::operator==(GFlowValue const& other) const {
  return kind == other.kind && gridsize == other.gridsize;
}

bool GFlowValue::operator!=(GFlowValue const& other) const {
  return kind != other.kind || gridsize != other.gridsize;
}

static GFieldSpecValues flowvalues_{
  {"column", GFlow::Column},
  {"row",    GFlow::Row},
  {"grid",   GFlow::Row, true},
  {"pie",    GFlow::Pie},
  {"stack",  GFlow::Stack}
};

static GPropSpec flowspec_{
  new GListFieldSpec<GFlowValue,int8_t>(&GFlowValue::kind, flowvalues_),
  new GNumFieldSpec<GFlowValue,int16_t>(&GFlowValue::gridsize)
};

bool gconvert(GString const& from, GFlowValue& to) {
  to.kind = GFlow::Row;
  to.gridsize = 0;
  int n = flowspec_.toValue(from, &to);
  if (n <= 0) gerror("Flow","invalid keyword or value: " + from);
  return n > 0;
}

bool gconvert(GFlowValue const& from, GString& to) {
  to = flowspec_.toString(&from);
  return true;
}

// - - - - - - 

GFlow::GFlow() {}

GFlow::GFlow(Kind kind, int16_t gridsize) {
  pmodes_.isSet = 1;
  value_.kind = kind;
  value_.gridsize = gridsize;
}

GFlow::GFlow(GFlowValue const& from) : Super(from) {}

GFlow::GFlow(GFlow const& from) : Super(from.value_) {}

GFlow::GFlow(GString const& value) {set(value);}

GFlow* GFlow::clone(bool copy_value) const {
  return copy_value ? new GFlow(*this) : new GFlow;
}

bool GFlow::onAdd(Gadget* g) {
  if (g) g->updateLayout();
  return true;
}

void GFlow::addModes(GadgetModes& m) {
  m.dir = (unsigned int)value_.kind;
}

void GFlow::updateGadget(Gadget& g) {
  g.modes_.dir = (unsigned int)value_.kind;
  g.updateLayout();
}

void GFlow::apply(Gadget* g, GRender& r, Specif specif) {
  if (specif < r.specif[GPropType::Flow]) return;
  r.specif[GPropType::Flow] = specif;

  r.flow_ = value_;
  if (g) g->modes_.dir = (unsigned int)value_.kind;
}

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

/*
GFlexValue::GFlexValue()
//: flex1(-1), flex2(-1), mode(GFlex::Undef) {}
: flex1(0), flex2(0), mode(GFlex::WidthHeight) {}
*/

GFlexValue::GFlexValue(int8_t flex1, int8_t flex2, int8_t mode)
: flex1(flex1), flex2(flex2), mode(mode) {}

bool GFlexValue::operator==(const GFlexValue& other) const {
  return flex1 == other.flex1 && flex2 == other.flex2 && mode == other.mode;
}

bool GFlexValue::operator!=(const GFlexValue& other) const {
  return flex1 != other.flex1 || flex2 != other.flex2 || mode != other.mode;
}

static GFieldSpecValues flexvalues_{
  {"self-wh", GFlex::SelfWH},
  {"self-flow",    GFlex::SelfFlow},
  {"parent-flow",  GFlex::ParentFlow},
};

static GPropSpec flexspec_{
  new GNumFieldSpec<GFlexValue,int8_t>(&GFlexValue::flex1),
  new GNumFieldSpec<GFlexValue,int8_t>(&GFlexValue::flex2),
  new GListFieldSpec<GFlexValue,int8_t>(&GFlexValue::mode, flexvalues_)
};

bool gconvert(GString const& from, GFlexValue& to) {
  to = {};
  //to.flex1 = to.flex2 = -1;
  //to.mode = GFlex::Undef;
  int n = flexspec_.toValue(from, &to);
  if (n <= 0) gerror("Flex","invalid keyword or value: " + from);
  return n > 0;
}

bool gconvert(GFlexValue const& from, GString& to) {
  to = flexspec_.toString(&from);
  return true;
}

// - - - - - -

GFlex::GFlex() = default;

GFlex::GFlex(int8_t flex1, int8_t flex2, Mode mode) {
  pmodes_.isSet = 1;
  value_.flex1 = flex1;
  value_.flex2 = flex2;
  value_.mode = mode;
}

GFlex::GFlex(GFlexValue const& value) : Super(value) {}

GFlex::GFlex(GFlex const& from) : Super(from.value_) {}

GFlex::GFlex(GString const& format) {set(format);}

GFlex* GFlex::clone(bool copy_value) const {
  return copy_value ? new GFlex(*this) : new GFlex;
}

void GFlex::set(int8_t flex1, int8_t flex2, Mode mode)  {
  set(GFlexValue(flex1, flex2, mode), false);
}

void GFlex::updateGadget(Gadget& g) {}

bool GFlex::onAdd(Gadget* g) {
  if (!g) return true;
  g->updateLayout();
  return true;
}

void GFlex::onRemove(Gadget* g) {
  if (!g) return;
  g->updateLayout();
}

void GFlex::apply(Gadget* g, GRender& r, Specif specif) {
  if (specif < r.specif[GPropType::Flex]) return;
  r.specif[GPropType::Flex] = specif;

  // note : dans gadget.cpp :
  // flex < 0 means flex must not be changed

  r.flex_ = value_;
    /*
    if (value_.flex1 >= 0) {
      r.flex_.flex1 = value_.flex1;
      r.flex_.mode = value_.mode;  // pas tres coherent
    }
    if (value_.flex2 >= 0) {
      r.flex_.flex2 = value_.flex1;
      r.flex_.mode = value_.mode;
    }*/
}

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

GAlignValue::GAlignValue() :
main(GAlign::Start), cross(GAlign::Start), icon(GAlign::Start) {}

GAlignValue::GAlignValue(int8_t main, int8_t cross, int8_t icon) :
main(main), cross(cross), icon(icon) {}

bool GAlignValue::operator==(const GAlignValue& other) const {
  return main == other.main && cross == other.cross && icon == other.icon;
}

bool GAlignValue::operator!=(const GAlignValue& other) const {
  return main != other.main || cross != other.cross || icon != other.icon;
}

static GFieldSpecValues alignvalues_{
  {"s",      GAlign::Start, true},
  {"c",      GAlign::Center, true},
  {"e",      GAlign::End, true},
  {"x",      GAlign::Extend, true},
  {"start",  GAlign::Start},
  {"center", GAlign::Center},
  {"end",    GAlign::End},
  {"extend", GAlign::Extend},
  {"stretch",GAlign::Extend, true},
  {"equal",  GAlign::Equal},
  {"eq",     GAlign::Equal, true},
};

static GPropSpec alignspec_{
  new GListFieldSpec<GAlignValue,int8_t>(&GAlignValue::main, alignvalues_),
  new GListFieldSpec<GAlignValue,int8_t>(&GAlignValue::cross, alignvalues_),
  new GListFieldSpec<GAlignValue,int8_t>(&GAlignValue::icon, alignvalues_)
};

bool gconvert(GString const& from, GAlignValue& to) {
  to.main  = GAlign::Start;
  to.cross = GAlign::Start;
  to.icon  = GAlign::Start;
  int n = alignspec_.toValue(from, &to);
  if (n <= 0) gerror("Align","invalid keyword or value: " + from);
  return n > 0;
}

bool gconvert(GAlignValue const& from, GString& to) {
  to = alignspec_.toString(&from);
  return true;
}

// - - - - - -

GAlign::GAlign() = default;

GAlign::GAlign(Layout main_dir, Layout cross_dir, Layout icon) {
  pmodes_.isSet = 1;
  value_.main = main_dir;
  value_.cross = cross_dir;
  value_.icon = icon;
}

GAlign:: GAlign(GAlign const& from) : Super(from.value_) {}

GAlign:: GAlign(GAlignValue const& value) : Super(value) {}

GAlign::GAlign(GString const& format) {set(format);}

GAlign* GAlign::clone(bool copy_value) const {
  return copy_value ? new GAlign(*this) : new GAlign;
}

void GAlign::setMainLayout(Layout val) {
  set(GAlignValue(val, value_.cross, value_.icon),false);
}

void GAlign::setCrossLaoyut(Layout val) {
  set(GAlignValue(value_.main, val, value_.icon),false);
}

void GAlign::setIconLayout(Layout val) {
  set(GAlignValue(value_.main, value_.cross,  val),false);
}

void GAlign::updateGadget(Gadget& g) {
  g.updateLayout();
}

void GAlign::apply(Gadget*, GRender& r, Specif specif) {
  if (specif < r.specif[GPropType::Align]) return;
  r.specif[GPropType::Align] = specif;

  r.align_ = value_;
}

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

GGapValue::GGapValue(float mainspacing, float crossspacing)
: mainspacing(mainspacing), crossspacing(crossspacing) {}

bool GGapValue::operator==(GGapValue const& from) const {
  return mainspacing == from.mainspacing && crossspacing == from.crossspacing;
}

bool GGapValue::operator!=(GGapValue const& from) const {
  return mainspacing != from.mainspacing || crossspacing != from.crossspacing;
}

static GPropSpec gapspec_{
  new GNumFieldSpec<GGapValue,float>(&GGapValue::mainspacing),
  new GNumFieldSpec<GGapValue,float>(&GGapValue::crossspacing)
};

bool gconvert(GString const& from, GGapValue& to) {
  to.mainspacing = 1.f;
  to.crossspacing = 1.f;
  int n = gapspec_.toValue(from, &to);
  if (n <= 0) gerror("Gap","invalid keyword or value: " + from);
  return n > 0;
}

bool gconvert(GGapValue const& from, GString& to) {
  to = gapspec_.toString(&from);
  return true;
}

// - - - -

GGap::GGap() = default;

GGap::GGap(float mainspacing, float crossspacing) {
  pmodes_.isSet = 1;
  value_.mainspacing = mainspacing;
  value_.crossspacing = crossspacing;
}

GGap::GGap(GGap const& from) : Super(from.value_) {}

GGap::GGap(GGapValue const& value) : Super(value) {}

GGap::GGap(GString const& format) {set(format);}

GGap* GGap::clone(bool copy_value) const {
  return copy_value ? new GGap(*this) : new GGap;
}

void GGap::updateGadget(Gadget& g) {
  g.updateLayout();
}

void GGap::apply(Gadget*, GRender& r, Specif specif) {
  if (specif < r.specif[GPropType::Gap]) return;
  r.specif[GPropType::Gap] = specif;

  r.gap_ = value_;
}

}
