//
// Modes of a gadget, control whether a gadget is enabled, shown, checked, etc.
//
// Guit GUI Toolkit (https://www.telecom-paris.fr/~elc/guit/)
// Copyright 2020 Eric Lecolinet (eric.lecolinet@gmail.com) - Telecom Paris
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//

#ifndef Guit_Modes_hpp
#define Guit_Modes_hpp
#include <gbool.hpp>
#include <gproplist.hpp>
namespace guit {

/// Prop that enables/disables a gadget.
class GEnable : public GBool {
public:
  GEnable();
  GEnable(bool value);
  GEnable(GString const& value);
  GEnable* clone(bool copy_value) const override;

  /// Future value in an Active Expression (see GExpr).
  auto operator~() {return GNumPropRef_<GEnable>(*this);}

  GEnable& operator=(bool value);
  GEnable& operator=(const GBool&);
  GEnable& operator=(GString const& value);

  using GType = GPropType_<GEnable,GBool>;
  static GType& Type();
  GType& type() const override {return Type();}

  void apply(Gadget*, GRender&, Specif) override;
  void updateGadget(Gadget&) override;
};

/// Returns a GEnable created in the heap.
/// @{
inline GEnable& Enable() {return *new GEnable();}
inline GEnable& Enable(bool value) {return *new GEnable(value);}
/// @}



/// Prop that shows/hides a gadget.
class GShow : public GBool {
public:
  GShow();
  GShow(bool value);
  GShow(GString const& value);
  GShow* clone(bool copy_value) const override;

  /// Future value in an Active Expression (see GExpr).
  auto operator~() {return GNumPropRef_<GShow>(*this);}

  GShow& operator=(bool value);
  GShow& operator=(const GBool& value);
  GShow& operator=(GString const& value);

  using GType = GPropType_<GShow,GBool>;
  static GType& Type();
  GType& type() const override {return Type();}

  void apply(Gadget*, GRender&, Specif) override;
  void updateGadget(Gadget&) override;
};

/// Returns a GShow created in the heap.
/// @{
inline GShow& Show() {return *new GShow();}
inline GShow& Show(bool value) {return *new GShow(value);}
/// @}


/// Prop that checks/unchecks a gadget.
/// Adding this prop makes the gadget selectable.
class GCheck : public GBool {
public:
  GCheck();
  GCheck(bool value);
  GCheck(GString const& value);
  GCheck* clone(bool copy_value) const override;

  /// Future value in an Active Expression (see GExpr).
  auto operator~() {return GNumPropRef_<GCheck>(*this);}

  GCheck& operator=(bool value);
  GCheck& operator=(GBool const& prop);
  GCheck& operator=(GString const& value);

  using GType = GPropType_<GCheck,GBool>;
  static GType& Type();
  GType& type() const override {return Type();}

  /// @internal.
  /// @{
  void apply(Gadget*, GRender&, Specif) override;
  void updateGadget(Gadget&) override;
  void addModes(GadgetModes&) override;
  void removeModes(GadgetModes&) override;
  /// @}
};

/// Returns a GCheck created in the heap.
/// @{
inline GCheck& Check() {return *new GCheck();}
inline GCheck& Check(bool value) {return *new GCheck(value);}
/// @}


/// Prop that allows setting misc modes in CSS files.
class GModes : public GProp {
public:
  GModes(GString const& modes = "");

  using GType = GPropType_<GModes,GProp>;
  static GType& Type();
  GType& type() const override {return Type();}

  void addModes(GadgetModes&) override;
  void apply(Gadget*, GRender&, Specif) final;

private:
  void setModes(GadgetModes&);
  struct Mode{uint64_t type; bool val;};
  std::vector<Mode> modes_;
};

}
#endif
