//
// internals: internal objects for notifying gadgets
//
// Guit GUI Toolkit (https://www.telecom-paris.fr/~elc/guit/)
// Copyright 2020 Eric Lecolinet (eric.lecolinet@gmail.com) - Telecom Paris
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//

#ifndef Guit_Notifier_hpp
#define Guit_Notifier_hpp
#include <gvarprop.hpp>
namespace guit {

// [Impl] Base clase of Notifiers.
class GNotifier : public GProp {
public:
  GNotifier* toNotifier() override {return this;}
  GString stringValue() const override;
  bool isEquivalent(const GProp&) const override {return false;}
  virtual void* receiver() = 0;

  static GPropType& Type();
  GPropType& type() const override {return Type();}
};


// [Impl] added to a sender to notify a receiver.
class GChangeNotifier : public GNotifier {
public:
  void fire(GEvent&) override {doSet();}
  virtual void* sender() = 0;
  virtual GVarProp* receiverAsProp() = 0;
  virtual void doSet() = 0;
  virtual bool doTest() = 0;
  bool recv_is_master_{};
  int8_t bindtag_{};  // aka GVarProp::BindTag
};


template <class S, class R>
class GChangePropNotifier : public GChangeNotifier {
public:
  GChangePropNotifier(S& sender, R& recv): sender_(sender), recv_(recv) {}
  
  void* sender() override {return &sender_;}
  void* receiver() override {return &recv_;}
  GVarProp* receiverAsProp() override {return &recv_;}
  void onRemove(Gadget*) override {recv_.removeNotifiers(&sender_);}
  
  void doSet() override {
    if (recv_.changing()) return;  // avoids oo loops because of cross refs
    recv_.changing(1);
    typename R::ValueType newval{};
    gconvert(sender_.value(), newval);
    //recv_.set(newval,false); //wrong: calls propagateChange instead of fireCallbacks!
    if (recv_.doChange(newval)) recv_.fireDefaultCallbacks();
    recv_.changing(0);
  }
  
  bool doTest() override {
    typename R::ValueType newval{};
    return gconvert(sender_.value(), newval) && recv_.isValid(newval);
  }
  
  S& sender_;
  R& recv_;
};


template <class S, class SF, class R, class RF>
class GChangeFieldNotifier : public GChangeNotifier {
public:
  GChangeFieldNotifier(GPropField<S,SF> const& sender, GPropField<R,RF> const& recv):
  sender_(sender), recv_(recv) {}
  
  void* sender() override {return &sender_;}
  void* receiver() override {return &(recv_.prop_);}
  GVarProp* receiverAsProp() override {return &(recv_.prop_);;}
  void onRemove(Gadget*) override {recv_.prop_.removeNotifiers(&(sender_.prop_));}
  
  void doSet() override {
    if (recv_.prop_.changing()) return;  // avoids oo loops because of crossrefs
    recv_.prop_.changing(1);
    auto oldfield = recv_.field_;
    gconvert(sender_.field_, recv_.field_);
    auto newval = recv_.prop_.value();
    recv_.field_ = oldfield;
    if (recv_.prop_.doChange(newval)) recv_.prop_.fireDefaultCallbacks();
    recv_.prop_.changing(0);
  }
  
  bool doTest() override{
    auto oldfield = recv_.field_;
    if (!gconvert(sender_.field_, recv_.field_)) {
      recv_.field_ = oldfield;
      return false;
    }
    auto newval = recv_.prop_.value();
    recv_.field_ = oldfield;
    return recv_.prop_.isValid(newval);
  }
  
  GPropField<S,SF> sender_;
  GPropField<R,RF> recv_;
};


class GExprNotifier : public GNotifier {
public:
  GExprNotifier(class GExprWrapper& wrap) : wrap(wrap) {}
  void* receiver() override {return &wrap;}
  void fire(GEvent&) override;
  class GExprWrapper& wrap;
};


class GTriggerNotifier : public GNotifier {
public:
  GTriggerNotifier(GObject& sender, class GExprTrigger& receiver);
  void* receiver() override {return &recv_;}
  class GExprTrigger& trigger() {return recv_;}
  void fire(GEvent& e) override;
  void onRemove(Gadget* g) override; 

  GObject& sender_;
  class GExprTrigger& recv_;
};


class GTriggerExprNotifier : public GTriggerNotifier {
public:
  GTriggerExprNotifier(GObject& sender, GExprTrigger& receiver, GBoolExpr& expr);
  void fire(GEvent&) override;
  GBoolExpr& expr_;
};

}
#endif
