//
// implementation of Number props
//
// Guit GUI Toolkit (https://www.telecom-paris.fr/~elc/guit/)
// Copyright 2020 Eric Lecolinet (eric.lecolinet@gmail.com) - Telecom Paris
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//

#include <limits>
#include <gint.hpp>
#include <gtext.hpp>
#include <gexpr.hpp>
#include <gbool.hpp>
#include <gint.hpp>
#include <gfloat.hpp>
#include <gcontrol.hpp>
using namespace std;
namespace guit {

GIMPLEMENT_PROP(Bool, GPropType::Bool)
GIMPLEMENT_PROP(Int, GPropType::Number)
GIMPLEMENT_PROP(Float, GPropType::Number)

GInt& Int(int value, int min, int max) {
  return *new GInt(value,min,max);
}

GFloat& Float(float value, float min, float max) {
  return *new GFloat(value,min,max);
}

GBool::GBool() {}

GBool::GBool(bool value) : Super(value) {}

GBool::GBool(GBool const& from) : Super(from.value_) {}

GBool::GBool(GString const& value) : GBool() {set(value);}

GBool* GBool::clone(bool copy_value) const {
  return copy_value ? new GBool(*this) : new GBool;
}

GBool& GBool::operator<<=(GBoolExpr& expr) {
  bindImpl(expr);
  this->set(bool(expr.floatValue()), false);
  return *this;
}

void GBool::set(const GProp& p) {
  set(p.boolValue(), false);
}

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

void GNumber::blend(const GProp& p1, const GProp& p2, float alpha)  {
  setF(alpha * p1.floatValue() + (1.f - alpha) * p2.floatValue());
}

void GNumber::updateGadget(Gadget& g) {
  g.update();
}

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

GInt::GInt() {}

GInt::GInt(int value) {doChange(value);}

GInt::GInt(int value, int min, int max) {
  setRange(min, max);
  doChange(value);  // will check isnide min, max
}

GInt::GInt(GInt const& from) {
  value_ = from.value_;
  setRange(from.min_, from.max_);
  pmodes_.autoReset = from.pmodes_.autoReset;
}

GInt::GInt(GString const& value) : GInt() {set(value);}

GInt* GInt::clone(bool copy_value) const {
  return copy_value ? new GInt(*this) : new GInt;
}

GInt& GInt::operator<<=(GNumExpr& expr) {
  bindImpl(expr);
  set( int(expr.floatValue()) , false);
  return *this;
}

GInt& GInt::setCycles(bool cycles) {
  pmodes_.autoReset = cycles;
  return *this;
}


void GInt::increase(int const& delta) {
  if (omodes_.immutable) immutableError("increase");
  else if (!pmodes_.autoReset) set(value_ + delta);
  else {
    if (delta > 0) {
      if (value_ <= max_ - delta) set(value_ + delta); else set(min_);
    }
    else if (delta < 0) {
      if (value_ >= min_ - delta) set(value_ + delta); else set(max_);
    }
  }
}

/*
 gpropset& GInt::selectAmong(const PropList& list) {
 auto* prop = new gpropsetValue(list);
 *this << On / [this,prop]{prop->select(value());};
 return *prop;
 }*/

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

GFloat::GFloat() {}

GFloat::GFloat(float value) {doChange(value);}

GFloat::GFloat(float value, float min, float max) {
  setRange(min, max);
  doChange(value);  // apres setRange() !
}

GFloat::GFloat(GFloat const& from) {
  value_ = from.value_;
  setRange(from.min_, from.max_);
}

GFloat::GFloat(GString const& value) : GFloat() {
  set(value);
}

GFloat* GFloat::clone(bool copy_value) const {
  return copy_value ? new GFloat(*this) : new GFloat;
}

GFloat& GFloat::operator<<=(GNumExpr& expr) {
  bindImpl(expr);
  this->set(expr.floatValue());
  return *this;
}

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

bool gconvert(const bool& from, GString& to) {
  to = from ? "true" : "false";
  return true;
}

bool gconvert(GString const& from, bool& to) {
  if (from=="no" || from=="false") {to = false; return true;}
  else if (from=="yes" || from=="true") {to = true; return true;}
  else {to = false; return false;}
}


// note: validity check needed by GTextField when linked to a GFloat
bool gconvert(GString const& from, float& to) {
  to = 0.f;                                      // accepts both . and , !
  if (from.empty() || from=="-" || from=="+" || from=="." || from==",") return true;
  /*
  std::size_t readcount = 0;
  try {
    to = std::stof(from, &readcount);    // current locale is used!
    return (readcount == from.size());
  }
  catch(...) {return false;}
   */
  char* pend{};
  to = std::strtof(from.data(), &pend);  // current locale is used!

  // str.data() == pend => error: no conversion could be performed
  // *pend != 0 => suffix after number
  return pend != from.data() && *pend == 0;
}

static void removeZeros(GString& to) {
  if (to.empty()) return;
  size_t dot = to.find_first_of(".,");
  if (dot == GString::npos) return;
  /* keep specified number of digits after .
   if (f && f->prec_ > 0) {
   if (dot + 1 + f->prec_ < to.size()) to.erase(dot+1+ f->prec_);
   return;
   }*/
  // remove ending 0 and . if nothing afterwards
  size_t pos = to.size()-1;
  while (pos >= 0) {
    if (to[pos] != '0') break;
    else --pos;
  }
  if (pos == dot) to.erase(pos);
  else to.erase(pos+1);
}

bool gconvert(const float& from, GString& to) {
  to = std::to_string(from);
  removeZeros(to);
  return true;
}

bool gconvert(const float& from, int& to) {
  to = int(from);
  if (from >= std::numeric_limits<int>::min() && from <= std::numeric_limits<int>::max()) return true;
  else {
    gerror("gconvert(float,int)", to_string(from)+" is too large to be converted into an int");
    return false;
  }
}

bool gconvert(const double& from, int& to) {
  to = int(from);
  if (from >= std::numeric_limits<int>::min() && from <= std::numeric_limits<int>::max()) return true;
  else {
    gerror("gconvert(double,int)", to_string(from)+" is too large to be converted into a float");
    return false;
  }
}

bool gconvert(const double& from, float& to) {
  to = float(from);
  if (from >= std::numeric_limits<int>::lowest() && from <= std::numeric_limits<int>::max()) return true;
  else {
    gerror("gconvert(double,float)", to_string(from)+" is too large to be converted into a float");
    return false;
  }
}

/*
bool gconvert(GString const& from, int& to) {
  to = 0;
  if (from.empty() || from=="-" || from=="+") return true;  // no hexa !!
   std::size_t readcount = 0;
   try {
   to = std::stoi(from, &readcount);
   return (readcount == from.size());
   }
   catch(...) {return false;}
}*/

// note: validity check used by GTextField when linked to an GInt
static bool conv_to_long(GString const& from, long& to) {
  to = 0;
  // just + or - is OK for GTextField
  if (from.empty() || from=="-" || from=="+") return true;

  char* pend{};
  to = std::strtol(from.data(), &pend, 10);  // current locale is used!

  // str.data() == pend => error: no conversion could be performed
  // *pend != 0 => suffix after number
  return pend != from.data() && *pend == 0;
}


bool gconvert(GString const& from, int& to) {
  long l{};
  bool stat = conv_to_long(from, l);
  to = (int)l;
  if (l >= std::numeric_limits<int>::min() && l <= std::numeric_limits<int>::max()) return stat;
  else {
    gerror("gconvert(string,int)", from+" is too large to be converted into an int");
    return false;
  }
}

bool gconvert(GString const& from, int16_t& to) {
  long l{};
  bool stat = conv_to_long(from, l);
  to = int16_t(l);
  if (l >= std::numeric_limits<int16_t>::min() && l <= std::numeric_limits<int16_t>::max()) return stat;
  else {
    gerror("gconvert(string,int16_t)", from+" is too large to be converted into an int16");
    return false;
  }
}

bool gconvert(GString const& from, int8_t& to) {
  long l{};
  bool stat = conv_to_long(from, l);
  to = (int8_t)l;
  if (l >= std::numeric_limits<int8_t>::min() && l <= std::numeric_limits<int8_t>::max()) return stat;
  else {
    gerror("gconvert(string,int8_t)", from+" is too large to be converted into an int8");
    return false;
  }
}

/*
 bool gconvert(const bool& from, int& to) {
 to = from ? 1 : 0;
 return true;
 }
 bool gconvert(const int& from, bool& to) {
 to = (from != 0);
 return true;
 }

bool gconvert(const double& from, int& to) {
  to = int(from);       // TEST SIZE!!
  return true;
}



bool gconvert(const double& from, GString& to) {
  to = std::to_string(from);      // current locale is used!
  return true;
}

bool gconvert(GString const& from, double& to) {
  to = 0.f;
  if (from.empty() || from=="-" || from=="+" || from=="." || from==",")
    return true;    // . or , !!
  std::size_t readcount = 0;
  try {
    to = std::stod(from, &readcount);    // current locale is used!
    return (readcount == from.size());
  }
  catch(...) {return false;}
}
 */

}
