//
// Position and z-layer of a Gadget
//
// Guit GUI Toolkit (https://www.telecom-paris.fr/~elc/guit/)
// Copyright 2020 Eric Lecolinet (eric.lecolinet@gmail.com) - Telecom Paris
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//

#ifndef GuitPos_hpp
#define GuitPos_hpp
#include <gvarprop.hpp>
namespace guit {

class GPosValue : public GPoint {
public:
  GPosValue() = default;
  GPosValue(float x, float y, int layer);
  GPosValue(GPoint const&, int layer);
  GPosValue& operator=(GPoint const&);
  bool operator==(GPosValue const&) const;
  bool operator!=(GPosValue const&) const;
  static GPosValue blend(GPosValue const& a, GPosValue const& b, float coeff);
  GString stringValue() const;
  int layer{};
  //uint8_t layer{};
};

inline bool gconvert(GPosValue const& from, GPosValue& to) {to = from; return true;}
bool gconvert(GString const& from, GPosValue& to);
bool gconvert(GPosValue const& from, GString& to);


/// Controls the position and layer of a Gadget.
/// - Gadgets having a GPos prop with a layer > 0 are displayed at the (x,y) position specified by this prop,
///  otherwise they are positioned automatically.
/// - The (x, y) position is the top-left position of the gadget (in pixels) relative to its parent. If the gadget
///  is a GWindow (or subclass), it is the postion on the screen.
/// - The layer specifies in which order gadgets are repainted inside their parent: gadgets with a higher
///  layer are painted on top of gadgets with a lower layer.
/// - 0 is the minimum layer value and specifies that the gadget is auto-positioned. The (x,y) position should
///  not be changed in that case, but can be read and will be the actual position of the gadget.
///
/// ## Binding and sharing GPos props
/// - A GPos can be bound / synced with other props (see gbind(), gsync()), so as it its fields (eg with
///  GSlider(s)): see xField(), yField(), layerField().
/// - A GPos can be shared by multiple gadgets, in which case their position are synced (except if the
///  layer is 0 for the above-explained reasons)
///
/// See also GRelatedPos (relative position from another Gadget) and GDgragPos (makes the
/// Gadget draggable interactively).
///
class GPos : public GVarProp_<GPosValue> {
public:
  using Super = GVarProp_<GPosValue>;
  using Layer = uint8_t;
  enum : Layer {MinLayer = 0, MaxLayer = 250};

  GPos();
  GPos(GString const& value);   ///< value = "x y layer" (layer is optional).
  GPos(float x, float y, Layer layer = 1);
  GPos(GPoint const& pos, Layer layer = 1);
  GPos(GPos const& value);
  GPos* clone(bool copy_value) const override;

  GPos* toPos() override {return this;}
  GPos const* toPos() const override {return this;}

  GPos& operator=(GPos const& pos) {set(pos); return *this;}
  GPos& operator=(GPoint const& pos) {set(pos); return *this;}
  GPos& operator=(GString const& pos) {set(pos); return *this;}
  
  using Super::set;
  void set(float x, float y);
  void set(float x, float y, Layer layer);
  void set(GPoint const& pos);
  void set(GProp const&) override;

  virtual void blend(GPos const& a, GPos const& b, float mix);
  void blend(GProp const& a, GProp const& b, float mix) override;

  void setX(float);
  void setY(float);
  void setLayer(int);

  virtual GPoint const& pos() const {return value_;}
  virtual float x() const {return value_.x;}
  virtual float y() const {return value_.y;}
  virtual int layer() const {return value_.layer;}

  /// Returns a Field for active expressions or binding.
  /// This field can be bound to a prop or another prop field using <<=.
  /// A textual Active Expression (GExpr) can refer this field.
  /// @{
  auto xField() {return GNumPropField<GPos,float>(*this, value_.x);}
  auto yField() {return GNumPropField<GPos,float>(*this, value_.y);}
  auto layerField() {return GNumPropField<GPos,int>(*this, value_.layer);}
  /// @}

  /// Binds this prop to an active expression.
  GPos& operator<<=(GTextExpr&);

  using GType = GPropType_<GPos,GProp>;
  static GType& Type();
  GType& type() const override {return Type();}

  /// @internal
  /// @{
  bool onAdd(Gadget*) override;
  void onRemove(Gadget*) override;
  void updateGadget(Gadget&) override;
  void apply(Gadget*, GRender&, Specif) final;
  /// @}
};

/// Returns a GPos created in the heap.
/// @{
GPos& Pos();
GPos& Pos(GString const& value);
GPos& Pos(float x, float y, GPos::Layer layer = 1);
/// @}


/// Draggable position: Makes the Gadget interactively movable by dragging it.
class GDragPos : public GPos {
public:
  GDragPos(float x, float y, Layer layer = 1) : GPos(x,y,layer) {}
  GDragPos(GPoint const& pos = {}, Layer layer = 1) : GPos(pos,layer) {}
  GDragPos(GDragPos const&);
  GDragPos(GString const&);
  GDragPos* clone(bool copy_value) const override;

  void fixX(bool val = true) {fixx_ = val;}
  void fixY(bool val = true) {fixy_ = val;}
  bool isXFixed() const {return fixx_;}
  bool isYFixed() const {return fixy_;}

  using GType = GPropType_<GDragPos,GPos>;
  static GType& Type();
  GType& type() const override {return Type();}

  /// @internal.
  /// @{
  bool fixx_{}, fixy_{};
  bool onAdd(Gadget*) override;
  void onRemove(Gadget*) override;
  void fire(GEvent&) override;
  /// @}
};


/// Relative position from another Gadget.
class GRelativePos : public GPos {
public:
  GRelativePos(Gadget& related, float xoffset, float yoffset, Layer = 1);
  GRelativePos(Gadget& related, GPoint const& offset = {}, Layer = 1);
  GRelativePos(GRelativePos const&);
  GRelativePos(GString const& = "");
  ~GRelativePos();
  GRelativePos* clone(bool copy_value) const override;

  Gadget* relatedGadget() const {return related_;}
  GPoint const& relatedPos() const {return relatedpos_;}
  virtual void updatePos();

  using GType = GPropType_<GRelativePos,GPos>;
  static GType& Type();
  GType& type() const override {return Type();}
  
  /// @internal.
  /// @{
  gptr<Gadget> related_;
  gptr<GProp> cond_;
  GPoint relatedpos_;
  /// @}
};

}
#endif
