//
// States, Transitions and Statecharts
//
// Guit GUI Toolkit (https://www.telecom-paris.fr/~elc/guit/)
// Copyright 2020 Eric Lecolinet (eric.lecolinet@gmail.com) - Telecom Paris
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//

#ifndef GuitState_hpp
#define GuitState_hpp
#include <list>
#include <gbool.hpp>
#include <gcond.hpp>
namespace guit {

class GState;
class GSuperState;
class GStateChart;
class GStateTransition;

/// State of a state machine.
/// Transitions can be added to gadgets as follows:
/// @code
///   auto& btn = Button()
///   << (state1 >> state2) / "Go" / [this]{go();}
///   << (state2 >> state3) / "Wait"
///   << (state3 >> state1) / "Stop" / [this]{stop();};
/// @endcode
/// When adding a callback function without specifying a trigger (as in this example), the default on::action
/// trigger of the Gadget is used (e.g. the button is cliked). A more general form is:
/// @code
///    mygadget << trigger / (state1 >> state2);
///    mygadget << trigger / (state1 >> state2) / [this]{doAction();
/// @endcode
/// The right-hand part could contain one or several callback(s), conditional prop(s) or conditional
/// expression(s): see GCond.
/// @see GSuperState and GStateChart
class GState : public GBool {
public:
  ~GState();

  /// Creates a state.
  /// The first version adds the newly created state to a superstate or a state machine.
  /// The second version just creates the state, it should be added to a superstate or a state machine
  /// using GSuperState::operator<<().
  /// @{
  GState(GSuperState& parent, GString const& name = "");
  GState(GString const& name = "");
  /// @}

  /// a GState can't be cloned.
  GState* clone(bool) const override {return nullptr;}

  /// Reference that can be added to an Active Expression (see GExpr).
  /// This reference will hold the future value, which can be (state entered) or false (state exited)
  /// @note operator+() and operator-() can also be used as shortcuts:
  /// @code
  ///    auto& state = State(parent);
  ///    mygadget
  ///    << +state / "yes" / GColor::green   // when the state is entered
  ///    << -state / "no"  / GColor::red;    // when the state is exited
  /// @endcode
  /// The color of the gadget will be green when the state is entered and red when it is exited
  auto operator~() {return GPropRef_<GState>(*this);}

  /// Returns the state value, i.e. true if the state is entered and false if it is exited.
  /// @{
  bool   entered() const;
  float  floatValue() const override {return entered();}
  GString stringValue() const override;
  /// @}

  /// Returns the name of the state (or "" if no name was specified).
  GString const name() const;

  /// Returns the parent State.
  /// The parent state is either the StateChart or a hierarchical State (GSuperState).
  GSuperState* parent() const {return parent_;}

  /// Adds a transition from this state to state2.
  /// Allows creating a transition (state1 >> state2) in an expression such as:
  /// @code
  ///    mygadget << (state1 >> state2);
  ///    mygadget << (state1 >> state2) / [this]{doAction();}
  ///    mygadget << trigger / (state1 >> state2) / [this]{doAction();}
  /// @endcode
  GStateTransition& operator>>(GState& state2);
  GStateTransition& operator>>(GState* state2);

  using GBool::add;
  virtual void add(GStateTransition*);
  using GBool::remove;
  virtual void remove(GStateTransition*);
  virtual void removeTransitions();
  virtual void enableGadgets(bool enable);
  bool isEquivalent(const GProp&) const override {return false;}

  using GType = GPropType_<GState,GBool>;
  static GType& Type();
  GType& type() const override {return Type();}

  /// @internal
  /// @{
  GSuperState* parent_{};
  GString* name_{};
  std::list<gptr<GStateTransition>> transitions_;
  std::forward_list<Gadget*> gadgets_;
  virtual void fireState(bool on);
  /// @}
};

/// Returns a GState created in the heap; see GState.
GState& State(GSuperState& parent, GString const& name = "");


/// Returns a conditional statement that calls a callback function or activates a Prop when a state transition is performed.
/// @see GState and GCond.
/// @{
GCond& operator/(GStateTransition&, GFunction const& fun);
GCond& operator/(GStateTransition&, GString const& text);
GCond& operator/(GStateTransition&, GProp& prop);
GCond& operator/(GStateTransition&, GProp* prop);
/// @}

/// Returns a conditional statement that evaluates this state transition when the first argument is fired.
/// @see GState and GCond.
/// @{
GCond& operator/(GTrigger&,  GStateTransition&);
GCond& operator/(GBoolExpr&, GStateTransition&);
GCond& operator/(GCond&,     GStateTransition&);
/// @}

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

/// Hierarchical State.
class GSuperState : public GState {
public:
  ~GSuperState();

  /// Creates a hierarchical state (see GState).
  /// @{
  GSuperState(GSuperState& parent, GString const& name = "");
  GSuperState(GString const& name = "");
  /// @}

  /// Sets the initial state.
  void setInitial(GState&);
  void setInitial(GState*);

  /// Returns the current state (null if none).
  GState* currentState() const {return current_;}

  /// Adds a substate to this hierarchical state (or state machine).
  /// Returns the superstate
  /// @{
  GSuperState& operator<<(GState& substate);
  GSuperState& operator<<(GState* substate);
  /// @}

  /// Creates and adds a substate to this hierarchical state (or state machine).
  /// Returns the newly created state
  GState& addState(GString const& name = "");

  /// Removes all states.
  void removeStates();

  void enableGadgets(bool enable) override;

  using GType = GPropType_<GSuperState,GState>;
  static GType& Type();
  GType& type() const override {return Type();}

protected:
  friend class GStateTransition;
  virtual void setCurrent(GState*);
  void fireState(bool on) override;
  std::vector<gptr<GState>> states_;
  GState *current_{nullptr}, *initial_{nullptr};
};

/// Returns a GSuperState (hierarchical state) created in the heap; see GSuperState.
GSuperState& SuperState(GSuperState&, GString const& name = "");


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

/// Statechart.
/// @see GSuperState.
class GStateChart : public GSuperState {
public:
  GStateChart(GString const& name = "");

  /// Sets the initial state and starts the state machine.
  void start(GState& initial_state);

  /// Starts/stops the state machine.
  /// Note: Won't start if there is no initial state!
  /// @{
  void start() {run(true);}
  void stop()  {run(false);}
  void run(bool yes);
  /// @}

  using GType = GPropType_<GStateChart,GSuperState>;
  static GType& Type();
  GType& type() const override {return Type();}
};

/// Returns a GStateChart created in the heap; see GStateChart.
GStateChart& StateChart(GString const& name = "");


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

/// Transition between two States.
/// See GState.
class GStateTransition : public GObject {
public:
  using IsGadgetChild = bool;

  GStateTransition();
  GStateTransition(GState& state1, GState& state2);
  ~GStateTransition();

  GString typeName() const override {return "StateTransition";}

protected:
  friend class GState;
  friend class GCondTrans;
  GState *state1_, *state2_;
  class GCondTrans* condtrans_;
  virtual bool canFire() const;
  virtual void doFire(GEvent&);
};

}
#endif
