//
//  Timers
//
// Guit GUI Toolkit (https://www.telecom-paris.fr/~elc/guit/)
// Copyright 2020 Eric Lecolinet (eric.lecolinet@gmail.com) - Telecom Paris
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//

#ifndef Guit_Timer_hpp
#define Guit_Timer_hpp
#include <gint.hpp>
namespace guit {

/// Timer object.
/// A timer fires callbacks after a certain delay. A timer can fire callbacks one time, N times 
/// (each time after waiting for the specifed delay) or indefinitively.
///
/// A timer has a value (returned by value() or operator(), which are inherited from GIntt), which is
/// the number of the current repetition. This feature can be useful for animations.
///
/// A timer repeating indefinitively, or 3 times. doIt() is called every 1000ms:
/// @code
///    auto& timer1 = Timer(1000)   << [this]{doIt();};  // fired every 1000ms, indefinitively
///    auto& timer2 = Timer(1000,3) << [this]{doIt();};  // fired every 1000ms, 3 times
/// @endcode
/// When adding a callback function without specifying a trigger (as in this example), the default on::action
/// trigger is used, which means that the function is called after each repetition. The right-hand part could
/// contain one or several callback(s), conditional prop(s) or conditional expression(s): see GCond.
///
/// A One Shot timer is auto-deleted after the first repetition:
/// @code
///    auto& timer = Timer(1000, GTimer::OneShot) << [&]{doIt();};  // one-shot (auto-deleted)
/// @endcode
/// More precisely, the timer is auto-deleted only if it is not gptr-pointed elsewhere.
///
/// The on::done trigger allows executing callbacks after the last repetition:
/// @code
///    auto& timer = Timer(1000,3)
///    << [this]{doIt();}
///    << on::done / [this]{doDone();};
/// @endcode
///
/// A timer can have 'beat counters'. Below, timer.addBeat(var, n, max) returns a GInt that is increased
/// by 1 when timer has been fired 'n' times and that it is reset to 0 when it reaches 'max' (so that 'seconds'
/// will count the seconds, 'minut's the minutes and 'hours' the hours);
/// @code
///    auto &timer = Timer(1000)
///    auto& seconds = timer.addBeat(1, 60);
///    auto& minutes = timer.addBeat(60, 60);
///    auto& hours = timer.addBeat(60*60, 24);
///
///    auto& display = HBox()
///    << hours << minutes << seconds;   // displays each value in a TextField
///
///    // callback function that is called every minute
///    minutes << [&]{ std::cout << "minutes: " << minutes << std::endl; }
/// @endcode
/// @see GAnim for creating animations.
class GTimer : public GInt {
public:
  enum {Infinite = -1, OneShot = -2};

  GTimer(GTime delay = 0, int repetitions = Infinite);
  GTimer(GString const& format);
  ~GTimer();

  /// Starts or stops the timer.
  /// starts(true) sets the timer value to 0 (use reset() for not setting the value to 0).
  /// @{
  void start(bool yes = true);
  void stop() {start(false);}
  /// @}

  /// Restarts without setting the value to 0.
  void restart();

  /// changes the number of repetitions.
  GTimer& operator=(int value) {set(value); return *this;}
  GTimer& operator=(GInt const& value) {set(value); return *this;}
  GTimer& operator=(GString const& value) {set(value); return *this;}

  /// Adds a callback or a conditional prop or expression to the timer (see GCond).
  GTimer& operator<<(GCond& c) {add(c); return *this;}

  /// Adds a state transition to the timer (see GState).
  GTimer& operator<<(class GStateTransition&);

  /// Adds a beat counter to the timer.
  /// See GTimer doc for an example. Both versions return the 'beat', the only difference is that the
  /// first version creates the beat (a new GInt), while the second version uses a preexisting GInt.
  /// @{
  GInt& addBeat(int timer_count, int max_value);
  GInt& addBeat(GInt& beat, int timer_count, int max_value);
  /// @}

  bool isRunning() const {return running_;}

  GTime   delay() const {return duration_;}
  GTimer& setDelay(GTime);

  /// How many times the timer is fired.
  /// A value of -1 or GTimer::Infinite means that the timer is fired indefinitively.
  int     repetitions() const {return count_;}
  GTimer& setRepetitions(int);

  int waitingCalls() const {return waitingcalls_;}

  using GType = GPropType_<GTimer,GInt>;
  static GType& Type();
  GType& type() const override {return Type();}

  /// @internal
  /// Guit provides full access to all fields, but beware using them is at your own risk.
  /// @{
  virtual void timerCB();
  bool running_{};
  int count_{}, waitingcalls_{};
  GTime duration_{}, nexttime_{};
  class GNatTimer* nattimer_{};
  gprotect<GTimer> self_;
  // @}
};

/// Returns a GTimer created in the heap.
/// @{
GTimer& Timer();
GTimer& Timer(GTime delay = 0, int repetitions = GTimer::Infinite);
GTimer& Timer(GString const& value);
/// @}

}
#endif
